from __future__ import annotations
from typing import Type, TYPE_CHECKING
import ast

from whyqd.models import SchemaActionModel, MorphActionModel, CategoryActionModel

if TYPE_CHECKING:
    from whyqd.models import CategoryModel, FieldModel
    from whyqd.crosswalk.base import BaseSchemaAction, BaseMorphAction, BaseCategoryAction
    from whyqd.core import SchemaDefinition


class ScriptParser:
    """Parsing utility functions for all types of action scripts.

    Supports parsing and validation of any action script of the form:

        "ACTION > [term] < [term]"

    Where term fields are optional. Has no opinion on what the terms are, or whether they are nested.
    """

    ###################################################################################################
    ### ACTION UTILITIES
    ###################################################################################################

    def get_action_model(self, *, action: str) -> SchemaActionModel | MorphActionModel | CategoryActionModel | None:
        """Return a specific set of Action definitions in response to an Action name.

        Parameters
        ----------
        action: str
            An Action name.

        Returns
        -------
        SchemaActionModel, MorphActionModel, CategoryActionModel or None.
            For the requested Action name. Or None, if it doesn't exist.
        """
        from whyqd.crosswalk.actions import default_actions

        return next((da for da in default_actions if da.name == action.upper()), None)

    def get_action_from_script(
        self, *, script: str
    ) -> SchemaActionModel | MorphActionModel | CategoryActionModel | None:
        """Return the first action term from a script as its Model type.

        Parameters
        ----------
        script: str
            Action script, defined as "ACTION > TERM < TERM"

        Raises
        ------
        ValueError if not a valid ACTION.

        Returns
        -------
        SchemaActionModel, MorphActionModel, or CategoryActionModel.
        """
        # Get the action, where any of the `<` or `>` referenced terms may be absent.
        root = self.get_split_terms(script=script, by="<")
        # There must always be a first term and it must *always* be an ACTION. Everything else is optional.
        root = self.get_split_terms(script=root[0], by=">")
        action = self.get_action_model(action=root[0])
        if not action:
            raise ValueError(f"Term '{root[0]} is not a recognised ACTION.")
        return action

    def get_action_class(
        self, *, actn: SchemaActionModel | MorphActionModel | CategoryActionModel
    ) -> Type[BaseSchemaAction | BaseMorphAction | BaseCategoryAction]:
        """Return the ACTION class for an ACTION model.

        Parameters
        ----------
        action: SchemaActionModel, MorphActionModel, CategoryActionModel

        Returns
        -------
        class of Action
        """
        from whyqd.crosswalk.actions import actions

        return actions[actn.name]

    ###################################################################################################
    ### FIELD UTILITIES
    ###################################################################################################

    def get_schema_field(self, *, term: str, schema: SchemaDefinition | list[SchemaDefinition]) -> FieldModel:
        """Recover a field model from a string.

        Parameters
        ----------
        term: str
        schema: SchemaDefinition, or a list of SchemaDefinition comprising of the source and destination schemas.

        Raises
        ------
        ValueError if the field term is not recognised.

        Returns
        -------
        FieldModel
        """
        if not isinstance(schema, list):
            schema = [schema]
        field = None
        for s in schema:
            field = s.fields.get(name=term)
            if field:
                break
        if not field:
            e = f"Field name is not recognised from either of the source or destination schema fields ({term})."
            if len(schema) == 1:
                e = f"Field name is not recognised from the schema fields ({term})."
            raise ValueError(e)
        return field

    ###################################################################################################
    ### SCRIPT PARSING UTILITIES
    ###################################################################################################

    def generate_contents(self, *, text) -> list[tuple[int, str]]:
        """Generate parenthesized contents in string as pairs (level, contents).

        Parameters
        ----------
        text: str

        Returns
        -------
        Generator

        References
        ----------
        https://stackoverflow.com/a/4285211/295606
        """
        stack = []
        for i, c in enumerate(text):
            if c == "[":
                stack.append(i)
            elif c == "]" and stack:
                start = stack.pop()
                yield (len(stack), text[start + 1 : i])

    def get_split_terms(self, *, script: str, by: str, maxsplit: int = 1) -> list[str]:
        # https://docs.python.org/3/library/stdtypes.html#str.split
        # str.split(sep=None, maxsplit=-1)
        return [s.strip() for s in script.split(sep=by, maxsplit=maxsplit)]

    def get_literal(self, *, text: str) -> str:
        literal = text
        try:
            literal = ast.literal_eval(text)
        except (ValueError, TypeError):
            pass
        except SyntaxError:
            # `literal_eval` strips special characters, leading to syntax errors
            if text.startswith("'") and text.endswith("'"):
                literal = text[1:-1]
        return literal

    def get_listed_literal(self, *, text: str) -> list[str]:
        listed_literal = []
        for t in text.split(","):
            try:
                listed_literal.append(self.get_literal(text=t))
            except SyntaxError:
                if t:
                    listed_literal.append(t)
        return listed_literal

    def get_hexed_script(self, *, script: str, fields: list[CategoryModel | FieldModel]) -> str:
        """Replace field names with its hex. Ensures that any fruity characters literals don't cause havoc during
        parsing.

        Parameters
        ----------
        script: str
        fields: list of CategoryModel, FieldModel

        Returns
        -------
        str
        """
        # Going to sort these so the longest is first to avoid replacing partial matches
        # https://docs.python.org/3/howto/sorting.html
        # Bool category names need to be converted to strings
        for f in sorted(fields, key=lambda f: len(str(f.name)), reverse=True):
            if f"'{f.name}'" in script:
                script = script.replace(f"'{f.name}'", f"'{f.uuid.hex}'")
        return script



from __future__ import annotations
from typing import List, Union, Type, TYPE_CHECKING
from pathlib import Path
from datetime import date, datetime
import pandas as _pd
import modin.pandas as pd
import numpy as np
from pyarrow.parquet import ParquetFile
import pyarrow as pa
import re
import locale
import ast

from whyqd.parsers.core import CoreParser
from whyqd.models import ColumnModel, DataSourceModel, DataSourceAttributeModel
from whyqd.dtypes import MimeType, FieldType
from whyqd.config.ray_init import ray_start

try:
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
except locale.Error:
    # Readthedocs has a problem, but difficult to replicate
    locale.setlocale(locale.LC_ALL, "")

if TYPE_CHECKING:
    from whyqd.core import SchemaDefinition


class DataSourceParser:
    """Get, review and restructure tabular source data."""

    def __init__(self):
        self.core = CoreParser()
        self.DATE_FORMATS = {
            "date": {"fmt": ["%Y-%m-%d"], "txt": ["YYYY-MM-DD"]},
            "datetime": {
                "fmt": ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M:%S %Z%z"],
                "txt": ["YYYY-MM-DD hh:mm:ss", "YYYY-MM-DD hh:mm:ss UTC+0000"],
            },
            "year": {"fmt": ["%Y"], "txt": ["YYYY"]},
        }

    ###################################################################################################
    ### TABULAR DATA READERS AND WRITERS
    ###################################################################################################

    def read_excel(self, *, source: str | Path, **kwargs) -> dict[str, pd.DataFrame] | pd.DataFrame:
        # This will default to returning a dictionary of dataframes for each sheet
        # https://pandas.pydata.org/docs/reference/api/pandas.read_excel.html
        ray_start()
        if kwargs.get("sheet_name"):
            # Only going to return a single dataframe
            # Modin can't seem to read sheets correctly
            return pd.DataFrame(_pd.read_excel(source, **kwargs))
        df = pd.read_excel(source, **kwargs)
        keys = list(df.keys())
        for k in keys:
            if df[k].empty:
                del df[k]
        if len(df.keys()) == 1:
            df = df[keys[0]]
        return df

    def read_csv(self, *, source: str | Path, **kwargs) -> pd.DataFrame:
        # New in pandas 1.3: will ignore encoding errors - perfect for this initial wrangling process
        # https://pandas.pydata.org/docs/reference/api/pandas.read_csv.html
        ray_start()
        kwargs["encoding_errors"] = "ignore"
        # Supposed to help with fruity separater guessing
        kwargs["engine"] = "python"
        if not kwargs.get("nrows"):
            df = pd.read_csv(source, **kwargs)
        else:
            kwargs["iterator"] = True
            kwargs["chunksize"] = 10000
            df_iterator = pd.read_csv(source, **kwargs)
            df = pd.concat(df_iterator, ignore_index=True)
        return df

    def read_parquet(self, *, source: str | Path, nrows: int | None = None, **kwargs) -> pd.DataFrame:
        # https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html
        if nrows:
            # https://stackoverflow.com/a/69888274/295606
            pf = ParquetFile(str(source))
            pf = next(pf.iter_batches(batch_size=nrows))
            # https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#pyarrow.Table.to_pandas
            return pa.Table.from_batches([pf]).to_pandas(**kwargs)
        ray_start()
        if "engine" not in kwargs:
            kwargs["engine"] = "pyarrow"
        return pd.read_parquet(source, **kwargs)

    def read_feather(self, *, source: str | Path, **kwargs) -> pd.DataFrame:
        # https://pandas.pydata.org/docs/reference/api/pandas.read_feather.html
        ray_start()
        return pd.read_feather(source, **kwargs)

    def get(
        self,
        *,
        source: str | Path | DataSourceModel,
        mimetype: str | MimeType | None = None,
        header: int | None = 0,
        preserve: str | list[str] | bool = False,
        names: list[str] | None = None,
        nrows: int | None = None,
        sheet_name: str | None = None,
        directory: str | Path | None = None,
        **attributes,
    ) -> dict[str, pd.DataFrame] | pd.DataFrame | None:
        """Return a Pandas dataframe from a given source.

        Accepts default pandas parameters for Excel and CSV, but the objective is to preserve the source data with
        little data conversion outside of the data wrangling process.

        Parameters
        ----------
        source: str | Path | DataSourceModel
            Path to source data file.
        mimetype: str | MimeType
            **whyqd** supports reading from CSV, XLS, XLSX, Feather and Parquet files.
        header: int, default 0
            Row to use as the column headings. Default is 0.
        preserve: str or list of str, default None
            Column names where variable type guessing must be prevented and the original data preserved.
            Critical for foreign key references with weird formats, like integers with leading `0`.
            Only works for CSV and Excel.
        names: list of str, default None
            If the source data has no header row, explicitly pass a list of names - in the correct order - to address
            the data. Only works for CSV and Excel.
        nrows: int, default None
            A specified number of rows to return. For review, it is faster to load only a small number.
            Only works for CSV and Excel.
        sheet_name: str, default None
            If the source is a multi-sheet Excel file, and you know the sheet, you can load that specifically. If you
            need to apply parameters specific to each sheet, it's best to load individually. Default, load all.
        directory: str | Path, default None
            An optional working directory for use for remote source data. Default will be from `settings`.
        attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns
        -------
        DataFrame or dict of DataFrame
        """
        ray_start()
        kwargs = {}
        kwargs |= attributes
        if isinstance(source, DataSourceModel):
            # Additional attributes stored in the model, e.g. End-of-line errors require `quoting=csv.QUOTE_NONE`
            #   quoting: int, default None
            #     CSV errors in opening a source file can be fixed by referencing quotation and end-of-line errors using
            #     `quoting=csv.QUOTE_NONE` and `import csv`.
            if source.attributes:
                kwargs |= source.attributes.terms
            mimetype = source.mime
            preserve = source.preserve
            sheet_name = source.sheet_name
            header = source.header
            if source.names:
                names = source.names
            source = source.path
        if self.core.check_uri(source=source):
            # Modin currently does not work with remote files. It needs a downloaded file.
            # Might have something to do with the way it imports data?
            source = self.core.download_uri_source(source=source, directory=directory)
        if not self.core.check_source(source=source):
            e = f"Source at `{source}` not found."
            raise FileNotFoundError(e)
        # These, currently, don't accept any additional parameters.
        mimetype = self.get_mimetype(mimetype=mimetype)
        if mimetype in [MimeType.PRQ, MimeType.PARQUET]:
            return self.read_parquet(source=source, nrows=nrows, **kwargs)
        if mimetype in [MimeType.FTR, MimeType.FEATHER]:
            return self.read_feather(source=source, **kwargs)
        # If the dtypes have not been set, then ensure that any provided preserved columns remain untouched
        # i.e. no forcing of text to numbers
        # defaulting to `dtype = object` ...
        if preserve:
            if isinstance(preserve, bool):
                kwargs["dtype"] = pd.StringDtype()
                # kwargs["keep_default_na"] = False
                # kwargs["na_values"] = ""
            else:
                if not isinstance(preserve, list):
                    preserve = [preserve]
                # kwargs["dtype"] = {k: object for k in preserve}
                kwargs["dtype"] = {k: pd.StringDtype() for k in preserve}
        kwargs["header"] = header
        if names:
            # Preserve all rows
            kwargs["header"] = None
            kwargs["names"] = names
        if nrows:
            kwargs["nrows"] = nrows
        # Check mimetype
        if mimetype in [MimeType.XLS, MimeType.XLSX]:
            kwargs["sheet_name"] = sheet_name
            return self.read_excel(source=source, **kwargs)
        if mimetype == MimeType.CSV:
            return self.read_csv(source=source, **kwargs)
        # Unknown mimetype
        raise FileExistsError(f"Source at `{source}` does not have a recognised mimetype `{mimetype}`.")

    def set(self, *, df: pd.DataFrame, source: str | Path, mimetype: MimeType) -> None:
        """Save a Pandas dataframe from a given source.

        Parameters
        ----------
        source: str
            Source filename.
        mimetype: MimeType
            **whyqd** supports saving to CSV, XLS, XLSX, Feather and Parquet files.
        df: pd.DataFrame
        """
        if not isinstance(source, Path):
            source = Path(source)
        try:
            source = source.with_suffix(f".{mimetype.name}")
        except Exception:
            raise FileExistsError(f"Save mimetype not supported, `{mimetype}`.")
        if mimetype in [MimeType.PRQ, MimeType.PARQUET]:
            df.to_parquet(path=source, engine="pyarrow", index=False)
        if mimetype in [MimeType.FTR, MimeType.FEATHER]:
            df.to_feather(path=source)
        if mimetype in [MimeType.XLS, MimeType.XLSX]:
            df.to_excel(source, index=False)
        if mimetype == MimeType.CSV:
            df.to_csv(source, index=False)

    ###################################################################################################
    ### TABULAR DATA SCHEMA COERSION UTILITIES
    ###################################################################################################

    def coerce_to_schema(self, *, df: pd.DataFrame, schema: Type[SchemaDefinition]) -> pd.DataFrame:
        validate = {"matched": [], "unmatched": [], "coerced": []}
        columns = self.get_header_columns(df=df)
        for c in columns:
            prospect = schema.fields.get(name=c.name)
            if prospect:
                if c.dtype == prospect.dtype:
                    validate["matched"].append(c.name)
                else:
                    # Try coerce the column to the type
                    df[c.name] = self.coerce_column_to_dtype(column=df[c.name], coerce=prospect.dtype)
                    validate["coerced"].append(c.name)
                df[c.name] = df[c.name].astype(FieldType(prospect.dtype).astype)
            else:
                validate["unmatched"].append(c.name)
        if validate["unmatched"]:
            self.core.show_warning(
                f"{len(validate['unmatched'])} columns in Data not found in Schema. {validate['unmatched']}"
            )
        if validate["coerced"]:
            self.core.show_warning(
                f"{len(validate['coerced'])} columns in Data were coerced to appropriate dtypes in Schema. {validate['coerced']}"
            )
        return df

    def coerce_column_to_dtype(self, *, column: pd.Series, coerce: str) -> pd.Series:
        parser = {
            "date": self.parse_dates,
            "usdate": self.parse_usdates,
            "datetime": self.parse_dates,
            "year": self.parse_dates,
            "number": self.parse_float,
            "integer": self.parse_int,
            "boolean": self.parse_bool,
            "array": self.parse_string_list,
            "string": self.parse_string,
        }
        return column.apply(lambda x: parser[coerce](x))

    def validate_schema_coersion(self, *, df: pd.DataFrame, schema: Type[SchemaDefinition]) -> bool:
        """Returns the MimeType representation of of a submitted source datatype."""
        for field in schema.fields.get_all():
            # Tests conversion to type
            series = df[field.name].astype(FieldType(field.dtype).astype)
            # Tests each constraint
            if field.constraints:
                if field.constraints.category:
                    # https://python-reference.readthedocs.io/en/latest/docs/comprehensions/set_comprehension.html
                    constraints = {c.name for c in field.constraints.category}
                    difference = {
                        term for term in series.unique() if not (term is pd.NA or pd.isnull(term) or term is None)
                    } - constraints
                    if difference:
                        raise ValueError(
                            f"Terms in `{field.name}` not defined as a category constraint: `{list(difference)}`"
                        )
                if field.constraints.required and len(series) != series.count():
                    raise ValueError(
                        f"Required constraint in `{field.name}` has `{len(series) - series.count()}` null values in `{len(series)}` rows."
                    )
                if field.constraints.unique and series.count() != len(series.unique()):
                    raise ValueError(
                        f"Unique constraint in `{field.name}` has `{series.count() - len(series.unique())}` non-unique values in `{series.count()}` rows."
                    )
                if field.constraints.minimum and series.min() < field.constraints.minimum:
                    raise ValueError(
                        f"Minimum constraint in `{field.name}` has value `{series.min()}` < `{field.constraints.minimum}`"
                    )
                if field.constraints.maximum and series.max() > field.constraints.maximum:
                    raise ValueError(
                        f"Minimum constraint in `{field.name}` has value `{series.max()}` > `{field.constraints.maximum}`"
                    )
        return True

    ###################################################################################################
    ### DATA MODEL SCHEMA DERIVATION AND WRITER UTILITIES
    ###################################################################################################

    def derive_data_model(
        self,
        *,
        source: Path | str,
        mimetype: str | MimeType,
        sheet_name: str | None = None,
        header: int | list[int] | None = 0,
        **attributes,
    ) -> DataSourceModel | list[DataSourceModel]:
        """Derive a data model schema (or list) from a data source. All columns will be coerced to `string` type to
        preserve data quality even though this is far less efficient.

        Parameters
        ----------
        source: str
            Source filename.
        mimetype: str or MimeType
            Pandas can read a diversity of mimetypes. **whyqd** supports `xls`, `xlsx`, `csv`, `parquet` and `feather`.
        sheet_name: str, default None.
            If `mimetype` is a multi-sheeted Excel file, specify which one to use.
        header: int | list[int | None] | None = 0,
            Row (0-indexed) to use for the column labels of the parsed DataFrame. If there are multiple sheets, then
            a list of integers should be provided.
            If `header` is `None`, row 0 will be treated as values and a set of field names will be generated indexed
            to the number of data columns.
        attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns
        -------
        List of DataSourceModel, or DataSourceModel
        """
        response = []
        mimetype = self.get_mimetype(mimetype=mimetype)
        sheet_names = [sheet_name]
        if sheet_name is None and mimetype in [MimeType.XLS, MimeType.XLSX]:
            # Check for multiple sheets
            df = self.read_excel(source=source, sheet_name=sheet_name, nrows=1, **attributes)
            if isinstance(df, dict):
                # Multiple sheets
                sheet_names = list(df.keys())
        for i, sheet_name in enumerate(sheet_names):
            header_i = header
            if isinstance(header, list):
                header_i = header[i]
            if header_i is None:
                header_names = self.create_header_names(
                    source=source, mimetype=mimetype, sheet_name=sheet_name, **attributes
                )
            else:
                header_names = self.get_header_names(
                    source=source, mimetype=mimetype, sheet_name=sheet_name, header=header_i, **attributes
                )
            response.append(
                self.get_source_data_model(
                    source=source,
                    mimetype=mimetype,
                    names=header_names,
                    sheet_name=sheet_name,
                    header=header_i,
                    **attributes,
                )
            )
        if len(response) > 1:
            return response
        return response[0]

    def get_source_data_model(
        self,
        *,
        df: pd.DataFrame | None = None,
        source: Path | str,
        mimetype: str | MimeType,
        names: list[str] | None = None,
        sheet_name: str | None = None,
        header: int | None = 0,
        **attributes,
    ) -> DataSourceModel:
        """Derive a data model schema from a SINGLE tabular data source. All columns will be coerced to `string` type to
        preserve data quality even though this is far less efficient.

        Parameters
        ----------
        source: str | Path
            Source filename.
        mimetype: str or MimeType
            Pandas can read a diversity of mimetypes. **whyqd** supports `xls`, `xlsx`, `csv`, `parquet` and `feather`.
        names: list of str
            A de-conflicted list of header-row names.
        sheet_name: str, default None.
            If `mimetype` is a multi-sheeted Excel file, specify which one to use.
        header: int, default 0
            Row (0-indexed) to use for the column labels of the parsed DataFrame.
        attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns
        -------
        DataSourceModel
        """
        del_df = False
        if df is None:
            del_df = True
            if header is None or header != 0:
                # If names are specified, then header is None
                df = self.get(
                    source=source,
                    mimetype=mimetype,
                    names=names,
                    preserve=names,
                    sheet_name=sheet_name,
                    header=None,
                    **attributes,
                )
            else:
                df = self.get(source=source, mimetype=mimetype, preserve=names, sheet_name=sheet_name, **attributes)
        columns = self.get_header_columns(df=df)
        preserve = [c.name for c in columns if c.dtype == "string"]
        source_data_model = {
            "path": str(source),
            "mime": self.get_mimetype(mimetype=mimetype),
            "sheet_name": sheet_name,
            "columns": columns,
            "preserve": preserve,
            "checksum": self.get_checksum(df=df),
            "header": header,
            "index": df.shape[0],
            "attributes": DataSourceAttributeModel.parse_obj(attributes),
        }
        if header is None or header != 0:
            source_data_model["names"] = names
        if del_df:
            del df
        return DataSourceModel(**source_data_model)

    def create_header_names(
        self,
        *,
        source: Path | str,
        mimetype: str | MimeType,
        sheet_name: str | None = None,
        nrows: int = 50,
        **attributes,
    ) -> list[str]:
        """
        Return a list of default column names - starting with `column-0` - to apply to a given SINGULAR tabular data
        source.

        Parameters
        ----------
        source: str
            Source filename.
        mimetype: MimeType, default MimeType.CSV
            Pandas can read a diversity of mimetypes, but whyqd has only been tested on `xls`, `xlsx` and `csv`.
        sheet_name: str, default None.
            If `mimetype` is a multi-sheeted Excel file, specify which one to use.
        nrows: int, default 50
            A specified number of rows to return. For review, it is faster to load only a small number.
        attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns
        -------
        List of str
        """
        df = self.get(source=source, mimetype=mimetype, sheet_name=sheet_name, nrows=nrows, **attributes)
        if (isinstance(df, dict) and not df) or df.empty:
            return []
        return [f"column_{i}" for i in range(len(df.columns))]

    def get_header_names(
        self,
        *,
        source: Path | str,
        mimetype: str | MimeType,
        sheet_name: str | None = None,
        header: int | None = 0,
        nrows: int = 1,
        **attributes,
    ) -> list[str]:
        """Return a list of column names to apply to a given SINGULAR tabular data source.

        Parameters
        ----------
        source: str
            Source filename.
        mimetype: MimeType, default MimeType.CSV
            Pandas can read a diversity of mimetypes, but whyqd has only been tested on `xls`, `xlsx` and `csv`.
        sheet_name: str, default None.
            If `mimetype` is a multi-sheeted Excel file, specify which one to use.
        header: int, default 0
            Row (0-indexed) to use for the column labels of the parsed DataFrame.
        nrows: int, default 1
            A specified number of rows to return. For review, it is faster to load only a small number.
        attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns
        -------
        List of str
        """
        if header is not None and header > nrows:
            nrows = header + 1
        df = self.get(source=source, mimetype=mimetype, sheet_name=sheet_name, header=header, nrows=nrows, **attributes)
        return df.columns.tolist()

    def get_header_columns(self, *, df: pd.DataFrame) -> List(ColumnModel):
        """Returns a list of ColumnModels from a source DataFrame.

        Parameters
        ----------
        df: pd.DataFrame
            Should be derived from `get` with a sensible default for `nrows` being 50.

        Returns
        -------
        List of ColumnModel
        """
        # Prepare summary
        try:
            columns = [
                {"name": k, "type": "number"}
                if v in ["float64", "int64", "Float64", "Int64"]
                else {"name": k, "type": "date"}
                if v in ["datetime64[ns]"]
                else {"name": k, "type": "string"}
                for k, v in df.dtypes.apply(lambda x: x.name).to_dict().items()
            ]
        except AttributeError:
            # No idea ... some seem to give shit
            columns = [
                {"name": k, "type": "number"}
                if v in ["float64", "int64", "Float64", "Int64"]
                else {"name": k, "type": "date"}
                if v in ["datetime64[ns]"]
                else {"name": k, "type": "string"}
                for k, v in df.dtypes.apply(lambda x: str(x)).to_dict().items()
            ]
        return [ColumnModel(**c) for c in columns]

    ###################################################################################################
    ### GENERAL UTILITIES
    ###################################################################################################

    def get_mimetype(self, mimetype: str | MimeType) -> MimeType:
        """Returns the MimeType representation of a submitted source datatype.

        Parameters
        ----------
        mimetype: str or MimeType

        Returns
        -------
        MimeType
        """
        if isinstance(mimetype, MimeType):
            return mimetype
        elif isinstance(mimetype, str):
            try:
                return MimeType.value_of(mimetype)
            except ValueError:
                return MimeType(mimetype)
        raise ValueError(f"MimeType not found for '{mimetype}'")

    def deduplicate_columns(self, *, df: pd.DataFrame) -> pd.Index:
        """
        Source: https://stackoverflow.com/a/65254771/295606
        Source: https://stackoverflow.com/a/55405151
        Returns a new column list permitting deduplication of dataframes which may result from merge.

        Parameters
        ----------
        df: pd.DataFrame

        Returns
        -------
        pd.Index
            Updated column names
        """
        column_index = pd.Series(df.columns.tolist())
        if df.columns.has_duplicates:
            duplicates = column_index[column_index.duplicated()].unique()
            for name in duplicates:
                dups = column_index == name
                replacements = [f"{name}{i}" if i != 0 else name for i in range(dups.sum())]
                column_index.loc[dups] = replacements
        return pd.Index(column_index)

    def get_checksum(self, *, df: pd.DataFrame, crosscheck: str = None) -> str:
        checksum = self.core.get_data_checksum(df=df)
        if crosscheck and checksum != crosscheck:
            raise ValueError(
                f"Checksum provided ({crosscheck}) is not equal to the generated checksum ({checksum}). Check your data source."
            )
        return checksum

    def check_date_format(self, *, date_type: str, date_value: str) -> bool:
        # https://stackoverflow.com/a/37045601
        # https://www.saltycrane.com/blog/2009/05/converting-time-zones-datetime-objects-python/
        for fmt in self.DATE_FORMATS[date_type]["fmt"]:
            try:
                if date_value == datetime.strptime(date_value, fmt).strftime(fmt):
                    return True
            except ValueError:
                continue
        raise ValueError(f"Incorrect date format, should be: `{self.DATE_FORMATS[date_type]['txt']}`")

    ###################################################################################################
    ### PANDAS TYPE PARSERS
    ###################################################################################################

    def _parse_dates(self, x: Union[None, str], dayfirst: bool = True) -> Union[pd.NaT, date.isoformat]:
        """
        This is the hard-won 'trust nobody', certainly not Americans, date parser. This requires that the user
        specify whether the day is first (ISO-style) or not (USA-style).

        NOTE: This is for dates, not datetimes, or times. If it is a properly formatted datetime, it might work
        but don't count on it.

        TODO: Replace with https://github.com/scrapinghub/dateparser
            The only concern is that dateparser.parse(x).date().isoformat() will coerce *any* string to a date,
            no matter *what* it is.

        Parameters
        ----------
        x: str | None
        dayfirst: bool
            Indicates whether this should be considered as an American-format date. Default is `True`, not American.

        Returns
        -------
        pd.NaT | date.isoformat
        """
        if pd.isnull(x):
            return pd.NaT
        if isinstance(x, str) and len(x) <= 10 and ":" in x:
            # This specific variation of a date is interpreted as a time, so ...
            x = re.sub(r"[\\/,\.:;]", "-", x)
        # Check if to_datetime can handle things
        if not pd.isnull(pd.to_datetime(x, errors="coerce", dayfirst=dayfirst)):
            return date.isoformat(pd.to_datetime(x, errors="coerce", dayfirst=dayfirst))
        # Manually see if coersion will work
        x = str(x).strip()[:10]
        # Handles separators in ["\\", "/", ".", "-", ":", ";", ","]
        x = re.sub(r"[\\/,\.:;]", "-", x)
        try:
            if dayfirst:
                # Some attempt at ISO
                y, m, d = x.split("-")
            else:
                # American
                m, d, y = x.split("-")
        except ValueError:
            return pd.NaT
        if dayfirst:
            # Order checks, ignoring US dates ... need to figure whether day or year is first
            if len(d) == 4 or (len(y) <= 2 and len(d) <= 2):
                # Swap the day and year positions
                d, m, y = x.split("-")
        # Fat finger on 1999 ... not going to check for other date errors as no way to figure out
        if len(y) == 4 and y[0] == "9":
            y = "1" + y[1:]
        x = f"{'0' + d if len(d) == 1 else d}-{'0' + m if len(m) == 1 else m}-{'0' + y if len(y) == 1 else y}"
        # Check if to_datetime can handle things
        try:
            if len(y) <= 2:
                # NOTE: if `len(y)` is two digits, then the "earliest" date is 1969 from `69`.
                # Other years will be the future ¯\_(ツ)_/¯
                x = datetime.strptime(x, "%d-%m-%y")
            else:
                x = datetime.strptime(x, "%d-%m-%Y")
        except ValueError:
            return pd.NaT
        x = date.isoformat(x)
        try:
            pd.Timestamp(x)
            return x
        except pd.errors.OutOfBoundsDatetime:
            return pd.NaT

    def parse_dates(self, x: Union[None, str]) -> Union[pd.NaT, date.isoformat]:
        return self._parse_dates(x, dayfirst=True)

    def parse_usdates(self, x: Union[None, str]) -> Union[pd.NaT, date.isoformat]:
        return self._parse_dates(x, dayfirst=False)

    def parse_dates_coerced(self, x: Union[None, str]) -> Union[pd.NaT, pd.Timestamp]:
        return pd.to_datetime(self._parse_dates(x, dayfirst=True), errors="coerce")

    def parse_usdates_coerced(self, x: Union[None, str]) -> Union[pd.NaT, pd.Timestamp]:
        return pd.to_datetime(self._parse_dates(x, dayfirst=False), errors="coerce")

    def _postprocess_parse_float(self, x: str) -> np.nan | float:
        """
        Regex to extract wrecked floats: https://stackoverflow.com/a/71206446
        Checked against: https://regex101.com/
        """
        # https://stackoverflow.com/a/71206446
        if x.group(3):
            return f"{x.group(1).replace(',','').replace('.','')}.{x.group(3)}"
        elif x.group(2):
            return f"{x.group(1).replace(',','').replace('.','')}"
        else:
            return np.nan

    def _preprocess_parse_float(self, x: str | int | float) -> np.nan | float:
        """
        Regex to extract wrecked floats: https://stackoverflow.com/a/385597
        Checked against: https://regex101.com/
        """
        try:
            return float(x)
        except ValueError:
            re_float = re.compile(
                r"""(?x)
            ^
                \D*     		# first, match an optional sign *and space*
                (             # then match integers or f.p. mantissas:
                    \d+       # start out with a ...
                    (
                        \.\d* # mantissa of the form a.b or a.
                    )?        # ? takes care of integers of the form a
                    |\.\d+     # mantissa of the form .b
                )
                ([eE][+-]?\d+)?  # finally, optionally match an exponent
            $"""
            )
            try:
                x = re_float.match(x).group(1)
                x = re.sub(r"[^e0-9,-\.]", "", str(x))
                return locale.atof(x)
            except (ValueError, AttributeError):
                return np.nan
        except TypeError:
            return np.nan

    def parse_float(self, x: str | int | float) -> np.nan | float:
        """
        Regex to extract wrecked floats
        Checked against: https://regex101.com/
        References:
            https://stackoverflow.com/a/73083844
            https://stackoverflow.com/a/385597
            https://stackoverflow.com/a/71206446
        """
        # 1. Try preprocess
        parsed = self._preprocess_parse_float(x)
        if np.isnan(parsed):
            # 2. Try postprocess and re-preprocess
            sign = ""
            try:
                # 2.1 Check for a leading negative sign
                re_float = r"[^\d+-/÷%\*]*"
                parsed = re.sub(re_float, "", x)
                if len(parsed) > 1 and parsed[0] in ["-"]:
                    sign = parsed[0]
            except (ValueError, AttributeError, TypeError):
                return np.nan
            # 2.2 Postprocess as comma/point-separated groups
            re_float = r"\b\d{1,2}\.\d{1,2}\.\d{2}(?:\d{2})?\b|\b(?<!\d[.,])(\d{1,3}(?=([.,])?)(?:\2\d{3})*|\d+)(?:(?(2)(?!\2))[.,](\d+))?\b(?![,.]\d)"
            parsed = list(filter(None, [self._postprocess_parse_float(x) for x in re.finditer(re_float, x)]))
            if len(parsed) == 1 and isinstance(parsed[0], str):
                parsed = sign + parsed[0]
                parsed = self._preprocess_parse_float(parsed)
            else:
                return np.nan
        return parsed

    def parse_int(self, x: str | int | float) -> np.nan | int:
        try:
            return int(str(x).split(".")[0])
        except ValueError:
            return None

    def parse_string(self, x: str) -> None | str:
        x = str(x).replace("\n", " ").replace("\r", "").replace("\t", "").replace("\\", "").strip()
        x = " ".join(str(x).split())
        if x and x[0] == "'":
            x = x[1:]
        if not x:
            return None
        return x

    def parse_bool(self, x: str) -> bool:
        if str(x).strip().lower() == "true":
            return True
        if str(x).strip().lower() == "false":
            return False
        return np.nan

    def parse_string_list(self, x: str) -> List[str]:
        """Coerce a column, which should contain a list of strings, from literal to actual."""
        if not isinstance(x, str):
            return x
        x = ast.literal_eval(x)
        x = [t.strip() for t in x]
        if not x:
            return None
        return x



from enum import Enum


class FieldType(str, Enum):
    STRING = "string"
    NUMBER = "number"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    ARRAY = "array"
    TIME = "time"
    USDATE = "usdate"
    DATE = "date"
    DATETIME = "datetime"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"
    ANY = "any"

    @property
    def describe(self):
        description = {
            "string": "Any text-based string.",
            "number": "Any number-based value, including integers and floats.",
            "integer": "Any integer-based value.",
            "boolean": "A boolean [true, false] value. Can set category constraints to fix term used.",
            "array": "Any valid array-based data.",
            "time": "Any time, with an optional date. Must be in ISO8601 format, hh:mm:ss.",
            "usdate": "Any American-formatted date without a time. Expected is MM-DD-YYYY, with any separator.",
            "date": "Any date without a time. Must be in ISO8601 format, YYYY-MM-DD.",
            "datetime": "Any date with a time. Must be in ISO8601 format, with UTC time specified (optionally) as YYYY-MM-DD hh:mm:ss Zz.",
            "month": "Any month, as month end frequency, formatted as YYYY-MM",
            "quarter": "Any quarter, as quarter end frequency, formatted as YYYY-MM.",
            "year": "Any year, as year end frequency, formatted as YYYY.",
            "any": "Any valid JSON data.",
        }
        return description[self.value]

    @property
    def astype(self):
        astypes = {
            "string": "string",
            "number": "Float64",
            "integer": "Int64",
            "boolean": "boolean",
            "array": "object",
            "time": "timedelta64[ns]",
            "date": "datetime64[ns]",
            "usdate": "datetime64[ns]",
            "datetime": "datetime64[ns, UTC]",
            "month": "period[M]",
            "quarter": "period[Q]",
            "year": "period[Y]",
            "any": "object",
        }
        return astypes[self.value]



from __future__ import annotations
from pathlib import Path

from whyqd.dtypes import MimeType
from whyqd.models import DataSourceModel, CrosswalkModel, TransformModel
from whyqd.core import CrosswalkDefinition
from whyqd.core.base import BaseDefinition
from whyqd.parsers import DataSourceParser
from whyqd.config.settings import settings


class TransformDefinition(BaseDefinition):
    """Create and manage a transform to perform a schema to schema crosswalk on a tabular data source.

    Parameters:
      transform: Path to a transform definition, or a dictionary conforming to a transform model.
      crosswalk: A definition, or a dictionary conforming to the CrosswalkModel, or a path to a saved definition.
      data_source: Path to a tabular data source, or a dictionary conforming to a data source model.

    Example:
      Create a new `TransformDefinition`, and perform a crosswalk, then save the definition and transformed data
      as follows:

      ```python
      import whyqd as qd

      transform = qd.TransformDefinition(crosswalk=CROSSWALK, data_source=DATASOURCE)
      transform.process()
      transform.save(directory=DIRECTORY)
      ```
    """

    def __init__(
        self,
        *,
        transform: TransformModel | dict | Path | str | None = None,
        crosswalk: CrosswalkDefinition | CrosswalkModel | dict | Path | str | None = None,
        data_source: DataSourceModel | dict | Path | str | None = None,
    ) -> None:
        super().__init__(model_name="transform")
        self.reader = DataSourceParser()
        self.data_source = None
        self.data = None
        self.crosswalk = CrosswalkDefinition()
        self.set(transform=transform, crosswalk=crosswalk, data_source=data_source)
        self.destination_mimetype = settings.WHYQD_DEFAULT_MIMETYPE
        self.destination_path = Path(settings.WHYQD_DIRECTORY)

    @property
    def get(self) -> TransformModel | None:
        """Get the transform model.

        Returns:
          Pydantic TransformModel or None
        """
        return self.model

    def set(
        self,
        *,
        transform: TransformModel | dict | Path | str | None = None,
        crosswalk: CrosswalkDefinition | CrosswalkModel | dict | Path | str | None = None,
        data_source: DataSourceModel | dict | Path | str | None = None,
    ) -> None:
        """Update or create the transform.

        Parameters:
          transform: Path to a transform definition, or a dictionary conforming to a transform model.
          crosswalk: A definition, or a dictionary conforming to the CrosswalkModel, or a path to a saved definition.
          data_source: Path to a tabular data source, or a dictionary conforming to a data source model.
        """
        self.model = self.core.create_or_update_model(modelType=TransformModel, source=transform, model=self.model)
        # And update the original data
        # https://fastapi.tiangolo.com/tutorial/body-updates/#partial-updates-with-patch
        if self.model.dataSource and not data_source:
            self.data_source = self.model.dataSource
        elif data_source:
            if isinstance(data_source, DataSourceModel):
                self.data_source = data_source
            else:
                self.data_source = self.core.create_or_update_model(modelType=DataSourceModel, source=data_source)
            self.model.dataSource = self.data_source
        if self.model.crosswalk and not crosswalk:
            self.crosswalk.set(crosswalk=self.model.crosswalk)
        elif crosswalk:
            if isinstance(crosswalk, CrosswalkDefinition):
                self.crosswalk = crosswalk
            else:
                self.crosswalk.set(crosswalk=crosswalk)
            self.model.crosswalk = self.crosswalk.get

    #########################################################################################
    # PERFORM TRANSFORMATION PROCESS
    #########################################################################################

    def process(self) -> None:
        """Perform a crosswalk. You can access the dataframe after completion at `.data`, if it exists.

        Raises:
          ValueError: If there are missing required destination fields in the crosswalk.
        """
        if not self.model.dataSource or not self.model.crosswalk:
            raise ValueError("Before performing data transform provide both a data source and a crosswalk.")
        # Validate crosswalk
        self.crosswalk.validate()
        # Perform crosswalk, coercing to source schema prior to crosswalk
        df = self.reader.get(source=self.model.dataSource)
        df = self.reader.coerce_to_schema(df=df, schema=self.crosswalk.schema_source)
        df = self.crosswalk.crud.transform_all(df=df)
        destination_names = [f.name for f in self.crosswalk.schema_destination.fields.get_all() if f.name in df.columns]
        # Coerce to schema and prepare data source model
        df = self.reader.coerce_to_schema(df=df[destination_names], schema=self.crosswalk.schema_destination)
        required_names = [f.name for f in self.crosswalk.schema_destination.fields.get_required()]
        if set(required_names) - set(df.columns):
            raise ValueError(
                f"Missing required destination fields in crosswalked data: {set(required_names) - set(df.columns)}"
            )
        self.data = df

    #########################################################################################
    # VALIDATION UTILITIES
    #########################################################################################

    def validate(
        self,
        *,
        transform: TransformModel | dict | Path | str,
        data_destination: DataSourceModel | dict | Path | str,
        mimetype_destination: str | MimeType | None = None,
        data_source: DataSourceModel | dict | Path | str | None = None,
        mimetype_source: str | MimeType | None = None,
    ) -> bool:
        """Validate the transformation process and all data checksums. Will perform all actions on each interim
        data source.

        Parameters:
          transform: Path to a transform definition, or a dictionary conforming to a transform model.
          data_destination: Path to a tabular data source, or a dictionary conforming to a data source model. Destination
                            data for crosswalk validation.
          mimetype_destination: **whyqd** supports reading from CSV, XLS, XLSX, Feather and Parquet files. Required if
                                `data_destination` is not of `DataSourceModel`.
          data_source: Path to a tabular data source, or a dictionary conforming to a data source model. Should be defined
                      in `transform`, but you may have a different version from a different location.
          mimetype_source: **whyqd** supports reading from CSV, XLS, XLSX, Feather and Parquet files. Required if
                          `data_source` is provided (i.e. not from the `transform`) and not of `DataSourceModel`.

        Raises:
          ValueError: If any steps fail to validate.

        Returns:
          A boolean `True` on successful validation.
        """
        # Prepare
        self.set(transform=transform)
        if data_source is not None:
            source_df = self.reader.get(
                source=data_source, mimetype=mimetype_source, preserve=self.model.dataSource.preserve
            )
        else:
            source_df = self.reader.get(source=self.model.dataSource)
        # Validate source
        if not self.model.dataSource.checksum:
            raise ValueError("Validation failed. Data source has no saved checksum.")
        self.reader.get_checksum(df=source_df, crosscheck=self.model.dataSource.checksum)
        # Validate destination
        destination_df = self.reader.get(
            source=data_destination, mimetype=mimetype_destination, preserve=self.model.dataDestination.preserve
        )
        if not self.model.dataDestination.checksum:
            raise ValueError("Validation failed. Data destination has no saved checksum.")
        self.reader.get_checksum(df=destination_df, crosscheck=self.model.dataDestination.checksum)
        # Validate crosswalk, coercing to source schema prior to crosswalk
        self.crosswalk.validate()
        source_df = self.reader.coerce_to_schema(df=source_df, schema=self.crosswalk.schema_source)
        crosswalk_df = self.crosswalk.crud.transform_all(df=source_df)
        destination_names = [
            f.name for f in self.crosswalk.schema_destination.fields.get_all() if f.name in crosswalk_df.columns
        ]
        crosswalk_df = self.reader.coerce_to_schema(
            df=crosswalk_df[destination_names], schema=self.crosswalk.schema_destination
        )
        required_names = [f.name for f in self.crosswalk.schema_destination.fields.get_required()]
        if set(required_names) - set(crosswalk_df.columns):
            raise ValueError(
                f"Validation failed. Missing required destination fields in crosswalked data: {set(required_names) - set(crosswalk_df.columns)}"
            )
        self.reader.get_checksum(df=crosswalk_df, crosscheck=self.model.dataDestination.checksum)
        return True

    #########################################################################################
    # SAVE UTILITIES
    #########################################################################################

    def save(
        self,
        *,
        filename: str | None = None,
        mimetype: str | None = None,
        directory: str | Path | None = None,
        created_by: str | None = None,
        hide_uuid: bool = False,
    ) -> bool:
        """Save model as a json file, and save crosswalked destination dataframe as a chosen mimetype.

        !!! info
            **NOTE:** by default, transformed data are saved as `PARQUET` as this is the most efficient.

            Declare your mime type like so:

            ```python
            MIMETYPE = "csv" # upper- or lower-case is fine
            ```


        Parameters:
          directory:  Defaults to working directory
          filename:  Defaults to model name
          mimetype: **whyqd** supports saving to CSV, XLS, XLSX, Feather and Parquet files. Defaults to Parquet.
          created_by:  Declare the model creator/updater
          hide_uuid:  Hide all UUIDs in the nested JSON output.

        Returns:
          Boolean True if saved.
        """
        if self.data is None:
            raise ValueError("No destination data available. First run a crosswalk.")
        if mimetype:
            mimetype = self.reader.get_mimetype(mimetype=mimetype)
            self.destination_mimetype = mimetype
        else:
            mimetype = self.reader.get_mimetype(mimetype=self.destination_mimetype)
        if directory and self.core.check_source(source=directory):
            self.destination_path = Path(directory)
        if filename:
            source = self.destination_path / f"{self.core.get_now()}-{filename}.{mimetype.name}"
        else:
            source = (
                self.destination_path
                / f"{self.core.get_now()}-{self.model.crosswalk.schemaDestination.name}.{mimetype.name}"
            )
        self.model.dataDestination = self.reader.get_source_data_model(df=self.data, source=source, mimetype=mimetype)
        self.reader.set(df=self.data, source=source, mimetype=mimetype)
        super().save(filename=filename, directory=directory, created_by=created_by, hide_uuid=hide_uuid)



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import FieldModel

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseSchemaAction):
    """
    Rename the source field to the destination field.

    !!! tip "Script template"
        ```python
        "RENAME > 'destination_field' < ['source_field']"
        ```

        Where only the first `source_field` will be used.

    This - ideally - is the most commonly-used script action as many source and destination fields have the same intention
    but different names.

    !!! example
        ```python
        "RENAME > 'occupant_name' < ['Primary Liable party name']"
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "RENAME"
        self.title = "Rename"
        self.description = (
            "Rename an existing field to conform to a schema name. Only valid where a single field is provided."
        )
        self.structure = [FieldModel]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: list[FieldModel],
    ) -> pd.DataFrame:
        # Rename, note, can only be one field if a rename ...
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rename.html
        # {'from': 'too'}
        return df.rename(index=str, columns={source[0].name: destination.name})



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import FieldModel

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseSchemaAction):
    """Use sparse data from a list of fields to populate a new field by iterating over a list of fields and selecting the
    next value in the list.

    !!! tip "Script template"
        ```python
        "SELECT > 'destination_field' < ['source_field', 'source_field', etc.]"
        ```

        Where order of `source_field` is important, each successive field in the list has priority over the ones before
        it (e.g. for columns A, B & C, values in C will have precedence over values in B and A). If there are nulls in
        A and B, but not B, then the returned value will be from B. If, however, there are values in A, B and C, then
        the returned value will be from C.

    !!! example
        ```python
        "SELECT > 'occupation_state_date' < ['Account Start date', 'Current Relief Award Start Date']"
        ```

        In this example, any dates in `Current Relief Award Start Date` will have precedence over `Account Start date`
        while nulls will be ignored, ensuring that `Account Start date` will be returned.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "SELECT"
        self.title = "Select"
        self.description = "Use sparse data from a list of fields to populate a new field. Order is important, each successive field in the list have priority over the ones before it (e.g. for columns A, B & C, values in C will have precedence over values in B and A)."
        self.structure = [FieldModel]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: list[FieldModel],
    ) -> pd.DataFrame:
        if destination.name not in df.columns:
            df[destination.name] = None
        for field in source:
            df[destination.name] = np.where(df[field.name].notnull(), df[field.name], df[destination.name])
        return df



from .core import CoreParser
from .datasource import DataSourceParser
from .script import ScriptParser
from .action import ActionParser
from .morph import MorphParser
from .category import CategoryParser



from __future__ import annotations
from typing import Union, Optional
from pydantic import BaseModel, Field, StrictBool
from uuid import UUID, uuid4


class CategoryModel(BaseModel):
    uuid: UUID = Field(default_factory=uuid4, description="Unique identity for the category. Automatically generated.")
    # Little gotcha ... if bool is 2nd, then any bools are automatically coerced to string ... order is priority.
    # Second gotcha, Pydantic is quite liberal with what it classifies as bool https://docs.pydantic.dev/latest/usage/types/booleans/
    name: Union[StrictBool, str] = Field(..., description="Unique category term.")
    description: Optional[str] = Field(None, description="Description of the unique term")



from .field import CRUDField
from .action import CRUDAction

# For a new basic set of CRUD operations you could just do

# from whyqd.metamodel.crud.base import CRUDBase
# from whyqd.metamodel.models import ItemModel

# item = CRUDBase[ItemModel]()



from __future__ import annotations
import modin.pandas as pd
import numpy as np

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import ModifierModel, FieldModel


class Action(BaseSchemaAction):
    """
    Create a new field by iterating over a list of fields and selecting the oldest value in the list.

    !!! tip "Script template"
        ```python
        "SELECT_OLDEST > 'destination_field' < ['source_field' + 'source_field_date', 'source_field' + 'source_field_date', etc.]"
        ```

        Where `+` links two columns together, explicitly declaring that the date in `source_field_date` is used to
        assign the order to the values in `source_field`. Unlike the `SELECT` action, `SELECT_OLDEST` takes its ordering from
        this date assignment.

    !!! example
        ```python
        "SELECT_OLDEST > 'occupation_state_date' < ['Current Relief Award Start Date' + 'Current Relief Award Start Date', 'Account Start date' + 'Account Start date']"
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "SELECT_OLDEST"
        self.title = "Select by oldest"
        self.description = "Use sparse data from a list of fields to populate a new field selecting by the oldest value. Field-pairs required, with the first containing the values, and the second the dates for comparison, linked by a '+' modifier (e.g. A+dA, B+dB, C+dC, values with the oldest associated date will have precedence over other values)."
        self.structure = [FieldModel, ModifierModel, FieldModel]

    @property
    def modifiers(self) -> list[ModifierModel]:
        return [ModifierModel(**{"name": "+", "title": "Links"})]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: list[FieldModel | ModifierModel],
    ) -> pd.DataFrame:
        base_date = None
        # Requires sets of 3 terms: field, +, date_field
        term_set = len(self.structure)
        for data, modifier, date in self.core.chunks(lst=source, n=term_set):
            if modifier.name != "+":
                raise ValueError(f"Field `{source}` has invalid SELECT_BY_OLD script. Please review.")
            if not base_date:
                # Start the comparison on the next round
                df.rename(index=str, columns={data.name: destination.name}, inplace=True)
                base_date = date.name
                if data.name == date.name:
                    # Just comparing date columns
                    base_date = destination.name
                df[base_date] = df[base_date].apply(self.reader.parse_dates_coerced)
                continue
            # np.where date is <> base_date and don't replace value with null
            # logic: if test_date not null & base_date <> test_date
            # False if (test_date == nan) | (base_date == nan) | base_date >< test_date
            # Therefore we need to test again for the alternatives
            df[date.name] = df[date.name].apply(self.reader.parse_dates_coerced)
            df[destination.name] = np.where(
                df[date.name].isnull() | (df[base_date] < df[date.name]),
                np.where(df[destination.name].notnull(), df[destination.name], df[data.name]),
                np.where(df[data.name].notnull(), df[data.name], df[destination.name]),
            )
            if base_date != destination.name:
                df[base_date] = np.where(
                    df[date.name].isnull() | (df[base_date] < df[date.name]),
                    np.where(df[base_date].notnull(), df[base_date], df[date.name]),
                    np.where(df[date.name].notnull(), df[date.name], df[base_date]),
                )
        return df



from __future__ import annotations
from pydantic import BaseModel, Field, validator, constr, AnyUrl
from typing import Optional
import math


class CitationModel(BaseModel):
    """BibTeX model for citations.

    https://www.bibtex.com/format/

    Includes some non-standard fields.
    """

    author: str = Field(
        ..., description="The name(s) of the author(s) (in the case of more than one author, separated by `and`)."
    )
    title: str = Field(..., description="The title of the work.")
    url: Optional[AnyUrl] = Field(
        None,
        description="The URL field is used to store the URL of a web page or FTP download. It is a non-standard BibTeX field.",
    )
    publisher: Optional[str] = Field(None, description="The publisher's name.")
    institution: Optional[str] = Field(
        None, description="The institution that was involved in the publishing, but not necessarily the publisher."
    )
    doi: Optional[str] = Field(
        None,
        description="The doi field is used to store the digital object identifier (DOI) of a journal article, conference paper, book chapter or book. It is a non-standard BibTeX field. It's recommended to simply use the DOI, and not a DOI link.",
    )
    month: Optional[constr(strip_whitespace=True, to_lower=True)] = Field(
        None,
        description="The month of publication (or, if unpublished, the month of creation). Use three-letter abbreviation.",
    )
    year: Optional[int] = Field(None, description="The year of publication (or, if unpublished, the year of creation).")
    licence: Optional[str] = Field(
        None, description="The terms under which the associated resource are licenced for reuse."
    )
    note: Optional[str] = Field(None, description="Miscellaneous extra information.")

    class Config:
        # anystr_strip_whitespace = True
        validate_assignment = True

    @validator("month")
    def month_abbreviation(cls, v):
        if v.lower() not in ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"]:
            raise ValueError("Month must conform to the BibTeX three-letter code definition.")
        return v.lower()

    @validator("year")
    def year_is_four(cls, v):
        if v < 1000 or int(math.log10(v)) + 1 != 4:
            raise ValueError(
                "Year must be in the period 1000 to 9999. There may be a year 9999 problem, but if you're still using this software by then, I'll be impressed."
            )
        return v



from __future__ import annotations
from typing import Type, TYPE_CHECKING
import modin.pandas as pd

from whyqd.parsers import CoreParser, ScriptParser
from whyqd.models import CategoryModel

if TYPE_CHECKING:
    from whyqd.models import FieldModel, CategoryActionModel
    from whyqd.core import SchemaDefinition
    from whyqd.crosswalk.base import BaseCategoryAction


class CategoryParser:
    """Parsing functions for category implementations of action scripts. These are support actions which must be run
    before `CATEGORISE`.

    This is a hybrid function:

    * *Get* a list of unique values from a source column, if categorising by unique. This step isn't necessary for
      boolean assignment,
    * `ASSIGN` subset of unique values (or all booleans) to a specified `CategoryModel` in a destination `FieldModel`.

    Scripts must be 'flat' and are of the forms::

        "ASSIGN_CATEGORY_UNIQUES > FieldModel::CategoryModel < FieldModel::[CategoryModel, etc.]"
        "ASSIGN_CATEGORY_UNIQUES > FieldModel::bool < FieldModel::[CategoryModel, etc.]"
        "ASSIGN_CATEGORY_BOOLEANS > FieldModel::bool < FieldModel"

    Where:

    * `FieldModel` is the destination column and the `::` linked `CategoryModel` defines what term the source values
      are to be assigned. This is defined in the `Schema`.
    * For `ASSIGN_CATEGORY_UNIQUES` the `list` of `CategoryModel` - unique values from `FieldModel` - will be assigned
      `::CategoryModel`.
    * For `ASSIGN_CATEGORY_BOOLEANS` values from the FieldModel are treated as boolean `True` or `False`, defined by `::bool`.
    """

    def __init__(self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None) -> None:
        """
        Parameters
        ----------
        schema_source: SchemaDefinition, optional
        schema_destination: SchemaDefinition, optional
        """
        self.core = CoreParser()
        self.parser = ScriptParser()
        self.schema_source = None
        self.schema_destination = None
        if schema_source and schema_destination:
            self.set_schema(schema_source=schema_source, schema_destination=schema_destination)

    def parse(
        self,
        *,
        script: str,
        action: BaseCategoryAction,
    ) -> dict[str, CategoryActionModel | FieldModel | CategoryModel | list[CategoryModel]]:
        """Generate the parsed dictionary of an initialised category action script.

        Parameters
        ----------
        script: str
            An action script.
        action: BaseCategoryAction

        Raises
        ------
        ValueError for any parsing errors.

        Returns
        -------
        dict
            Parsed dictionary of an initialised category action script.
        """
        # [CHANGE] Will hex both fields and category terms ...
        if not self.schema_source or not self.schema_destination:
            raise ValueError("Schema for both source and destination has not been provided.")
        # Too many edge cases ... cannot resolve this in a satisfactory way ... 
        # Going to have to hope text issues don't emerge
        # script = self.get_hexed_script(script=script)
        parsed = action.parse(script=script)
        # Validate the category script and get the required compone
        destination, category, source, assigned, unassigned = None, None, None, None, None
        # Get source and source category terms
        source = self.parser.get_literal(text=parsed["source"])
        source = self.get_schema_field(term=source, is_source=True)
        # Get category terms from the field if the action.structure requires it
        if action.structure and (not source.constraints or not source.constraints.category):
            raise ValueError(f"Source field ({source.name}) is invalid since it has no assigned category constraints.")
        if parsed.get("source_category"):
            if action.structure:
                all_uniques = source.constraints.category
                assigned = []
                failed = []
                assigned_uniques = self.get_assigned_uniques(text=parsed["source_category"])
                for c in assigned_uniques:
                    category = self.get_schema_field_category(field=source, term=c, is_source=True)
                    if category:
                        assigned.append(category)
                    elif not category and isinstance(c, bool) and len(assigned_uniques) == 1:
                        # Coercing a column to a boolean field based on presence or absence of values
                        # Can't assign more than True or False
                        # Create a temporary category
                        assigned.append(CategoryModel(**{"name": c}))
                    else:
                        failed.append(c)
                    if failed:
                        raise ValueError(f"Assigned category not found in source field categories {set(failed)}.")
                unassigned = [c for c in all_uniques if c.uuid.hex not in assigned_uniques]
            else:
                assigned = parsed["source_category"]
        # Get destination column and assigned category term
        destination = self.parser.get_literal(text=parsed["destination"])
        destination = self.schema_destination.fields.get(name=destination)
        if parsed.get("category"):
            parsed_category = self.parser.get_literal(text=parsed["category"])
            category = self.get_schema_field_category(field=destination, term=parsed_category, is_source=False)
        if not destination and category:
            raise ValueError(
                f"Destination field and category are not valid for this category action script ({parsed['destination']}, {parsed['category']})."
            )
        # Return validated terms
        return {
            "action": action,
            "destination": destination,
            "category": category,
            "source": source,
            "assigned": assigned,
            "unassigned": unassigned,
        }

    ###################################################################################################
    ### IMPLEMENT VALIDATED SCRIPT
    ###################################################################################################

    def transform(
        self,
        *,
        df: pd.DataFrame,
        action: Type[BaseCategoryAction],
        destination: FieldModel,
        source: FieldModel,
        category: CategoryModel = None,
        assigned: list[CategoryModel] | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Transform a dataframe according to a categorisation script.

        Parameters
        ----------
        df: DataFrame
            Working data to be transformed
        action: SchemaActionModel
        destination: FieldModel
        source: list of ModifierModel, and dicts of nested transforms, default None
        assigned: list of dict
            Specific to CATEGORISE actions. Each dict has values for: Assignment ACTION, destination schema field,
            schema category, source data column, and a list of source data column category terms assigned to that
            schema category.

        Returns
        -------
        Dataframe
            Containing the implementation of all nested transformations
        """
        return action.transform(df=df, destination=destination, source=source, category=category, assigned=assigned)

    ###################################################################################################
    ### SUPPORT UTILITIES
    ###################################################################################################

    def set_schema(
        self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None
    ) -> None:
        """Set SchemaDefinitions for the parser.

        Parameters
        ----------
        schema_source: SchemaDefinition
        schema_destination: SchemaDefinition
        """
        if not schema_source or not schema_destination:
            raise ValueError("Schema for both source and destination has not been provided.")
        self.schema_source = schema_source
        self.schema_destination = schema_destination

    def get_schema_field_category(self, *, field: FieldModel, term: str, is_source: bool = True) -> CategoryModel | None:
        """
        Recover a field category model from a string. It is possible that source and destination schema category share 
        a common name, so we need to check all possibilities.

        Parameters
        ----------
        field: FieldModel
        term: str
        is_source: bool
            Boolean True for source schema, False for destination schema

        Returns
        -------
        CategoryModel | None
        """
        if not field:
            return None
        schema = self.schema_destination
        alt_schema = self.schema_source
        if is_source:
            schema = self.schema_source
            alt_schema = self.schema_destination
        category = schema.fields.get_category(name=field.uuid.hex, category=term)
        # Disambiguation step ...
        if not category:
            # Any schema field
            for schema_field in schema.fields.get_all():
                if schema_field.constraints and schema_field.constraints.category:
                    category = schema.fields.get_category(name=schema_field.uuid.hex, category=term)
                    if category:
                        category = schema.fields.get_category(name=field.uuid.hex, category=category.name)
                        break
        if not category:
            # Any alt schema field
            for schema_field in alt_schema.fields.get_all():
                if schema_field.constraints and schema_field.constraints.category:
                    category = alt_schema.fields.get_category(name=schema_field.uuid.hex, category=term)
                    if category:
                        category = schema.fields.get_category(name=field.uuid.hex, category=category.name)
                        break
        return category

    def get_schema_field(self, *, term: str, is_source: bool = True) -> FieldModel:
        """
        Recover a field model from a string. It is possible that source and destination schema share a common name,
        so we need to check both, but the categorical terms must come from the correct field. Raise an error if we
        can't identify appropriately.

        Parameters
        ----------
        term: str
        is_source: bool
            Boolean True for source schema, False for destination schema

        Raises
        ------
        ValueError if the field term is not recognised from the appropriate schema.

        Returns
        -------
        FieldModel
        """
        schema = self.schema_destination
        alt_schema = self.schema_source
        if is_source:
            schema = self.schema_source
            alt_schema = self.schema_destination
        field = schema.fields.get(name=term)
        if not field:
            field = alt_schema.fields.get(name=term)
            field = schema.fields.get(name=field.name)
        if not field:
            s = "destination"
            if is_source:
                s = "source"
            raise ValueError(f"Field name is not recognised from the {s} schema fields ({term}).")
        return field

    def get_hexed_script(self, *, script: str) -> str:
        # Changes fields to uuid hexes
        all_fields = [field for s in [self.schema_source, self.schema_destination] for field in s.get.fields]
        all_categories = [
            category
            for f in all_fields
            if (f.constraints and f.constraints.category)
            for category in f.constraints.category
        ]
        script = self.parser.get_hexed_script(script=script, fields=all_fields + all_categories)
        return ",".join([s.strip() for s in script.split(",") if s.strip()])

    def get_assigned_uniques(self, *, text: str) -> list[str]:
        terms = list(self.parser.generate_contents(text=text))
        if len(terms) != 1:
            raise ValueError(f"Category assignment actions must not be nested. ({text}).")
        return [self.parser.get_literal(text=t) for t in self.parser.get_split_terms(script=terms[0][1], by=",", maxsplit=-1)]



from __future__ import annotations
import modin.pandas as pd
import numpy as np

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import ModifierModel, FieldModel


class Action(BaseSchemaAction):
    """Calculate the value of a field derived from the values of other fields. Requires a `MODIFIER` indicating whether
    the fields should be added or subtracted from the current total.

    !!! tip "Script template"
        ```python
        "CALCULATE > 'destination_field' < [modifier 'source_field', modifier 'source_field', etc]"
        ```

        Where the `modifier` is either of `+` or `-`. If you want to change the sign of the values in a column,
        *before* further arithmatic, you'll need to do this sort of script:

        ```python
        "CALCULATE > 'destination_field' < [modifier 'source_field', modifier CALCULATE < [- 'source_field']"
        ```

    !!! example
        ```python
        "CALCULATE > 'total' < [+ 'income', - 'expenses']"
        ```

    !!! warning
        This is still very rudimentary. Enhancements should stick to basic `reverse Polish notation` arithmetic.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "CALCULATE"
        self.title = "Calculate"
        self.description = "Derive a calculation from a list of fields. Each field must have a modifier, including the first (e.g. +A -B +C)."
        self.structure = [ModifierModel, FieldModel]

    @property
    def modifiers(self) -> list[ModifierModel]:
        return [
            ModifierModel(**{"name": "+", "title": "Add"}),
            ModifierModel(**{"name": "-", "title": "Subtract"}),
        ]

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: list[FieldModel | ModifierModel],
    ) -> pd.DataFrame:
        term_set = len(self.structure)
        add_fields = [
            field.name for modifier, field in self.core.chunks(lst=source, n=term_set) if modifier.name == "+"
        ]
        sub_fields = [
            field.name for modifier, field in self.core.chunks(lst=source, n=term_set) if modifier.name == "-"
        ]
        for field in add_fields + sub_fields:
            df[field] = df[field].apply(self.reader.parse_float)
        # Need to maintain NaNs ... default is to treat NaNs as zeros, so even a sum of two NaNs is zero
        # If we don't know, we don't know ... but ... if a sum is mixed, then ignore the NaNs
        _add = df[add_fields].sum(min_count=1, axis=1).array
        _sub = df[sub_fields].sum(min_count=1, axis=1).array
        # Need to use 'logical_and' since comparing NaNs is a disaster
        df[destination.name] = np.where(
            np.logical_and(np.isnan(_add), np.isnan(_sub)), np.nan, np.nan_to_num(_add) - np.nan_to_num(_sub)
        )
        return df



from __future__ import annotations
import modin.pandas as pd

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import FieldModel


class Action(BaseSchemaAction):
    """Pivot a list of columns to create a new long table format.

    !!! tip "Script template"
        ```python
        "PIVOT_LONGER > ['name_field', 'value_field'] < ['source_field', 'source_field', etc.]"
        ```

        Will assign to the destination fields as:

        - name_field: containing the original source field names, as appropriate,
        - value_field: containing the original values corresponding to the original source fields.

    !!! example
        ```python
        "PIVOT_LONGER > ['year', 'values'] < ['1990', '1991', '1992', '1993']"
        ```
    """

    def __init__(self):
        super().__init__()
        self.name = "PIVOT_LONGER"
        self.title = "Pivot longer"
        self.description = "Transform a DataFrame from wide to long format."
        self.structure = [FieldModel]

    def validate(self, *, destination: list[FieldModel], source: list) -> bool:
        if not isinstance(destination, list) and len(destination) == 2:
            raise ValueError(f"Action source script does not conform to required structure. ({destination})")
        return super().validate(destination=destination, source=source)

    def transform(self, *, df: pd.DataFrame, destination: list[FieldModel], source: list[FieldModel]) -> pd.DataFrame:
        columns = [c.name for c in source]
        id_columns = list(set(df.columns.values.tolist()).difference(set(columns)))
        name_field, value_name = destination[0].name, destination[1].name
        # https://pandas.pydata.org/docs/reference/api/pandas.melt.html
        return pd.melt(df, id_vars=id_columns, value_vars=columns, var_name=name_field, value_name=value_name)



import ray
import os
# import json
import shutil
from pathlib import Path

from whyqd.config.settings import settings


def clear_spillway() -> bool:
    # https://stackoverflow.com/a/56151260/295606
    # only do this if ray is not initialised, otherwise it'll crash. horribly.
    WHYQD_SPILLWAY = ray._private.utils.get_user_temp_dir()
    if not Path(WHYQD_SPILLWAY).exists():
        return False
    for path in Path(WHYQD_SPILLWAY).iterdir():
        try:
            if path.is_file():
                path.unlink()
            elif path.is_dir():
                shutil.rmtree(path)
        except Exception as e:
            print(f"Failed to delete {path}. Reason: {e}")
    return True


def ray_start(**kwargs):
    # https://docs.ray.io/en/latest/ray-core/starting-ray.html
    # https://docs.ray.io/en/latest/ray-core/configure.html
    # https://github.com/modin-project/modin/issues/3845
    # https://stackoverflow.com/questions/57765260/is-there-a-way-to-limit-ray-object-storage-max-memory-usage
    # 10**9 == 1gb ==> try 8gb?
    # https://docs.ray.io/en/latest/ray-core/objects/object-spilling.html
    # Spilling is a way of paging from memory onto the drive.
    # https://docs.docker.com/compose/compose-file/compose-file-v3/#shm_size
    # In Docker, can choose to use dev_shm as a faster access than tmp
    # dev_shm must be aligned in docker-compose, and note that it's not under `build`
    #   shm_size: "8gb"
    if not ray.is_initialized():
        # clear the spill path of any old artifacts
        clear_spillway()
        # initialise - kwargs allow overwriting of the settings (or ignoring the .env)
        kwargs["ignore_reinit_error"] = True
        kwargs["runtime_env"] = {"env_vars": {"__MODIN_AUTOIMPORT_PANDAS__": "1"}}
        if not kwargs.get("num_cpus") and settings.WHYQD_CPUS:
            kwargs["num_cpus"] = settings.WHYQD_CPUS
        if not kwargs.get("_memory") and settings.WHYQD_MEMORY:
            kwargs["_memory"] = settings.WHYQD_MEMORY
        if not kwargs.get("object_store_memory") and settings.WHYQD_MEMORY:
            kwargs["object_store_memory"] = settings.WHYQD_MEMORY
        if not kwargs.get("_temp_dir") and settings.WHYQD_SPILLWAY:
            # TODO: On some environments, I'm getting weird timeouts
            if not ray._private.utils.get_user_temp_dir().endswith(settings.WHYQD_SPILLWAY):
                kwargs["_temp_dir"] = f"{ray._private.utils.get_user_temp_dir()}{settings.WHYQD_SPILLWAY}"
            pass
        # if not kwargs.get("_system_config") and settings.WHYQD_SPILLWAY:
        #     kwargs["_system_config"] = {
        #         "object_spilling_config": json.dumps(
        #             {
        #                 "type": "filesystem",
        #                 "params": {
        #                     "directory_path": str(settings.WHYQD_SPILLWAY),
        #                 },
        #             },
        #         )
        #     }
        ray.init(**kwargs)
        os.environ["MODIN_ENGINE"] = "ray"
        os.environ["MODIN_OUT_OF_CORE"] = "true"
        if kwargs.get("_memory"):
            os.environ["MODIN_MEMORY"] = str(kwargs["_memory"])
        if kwargs.get("_temp_dir"):
            os.environ["RAY_TMPDIR"] = kwargs["_temp_dir"]


def ray_stop():
    if ray.is_initialized():
        ray.shutdown()


def ray_restart():
    ray_stop()
    ray_start()



from __future__ import annotations
from pydantic import BaseModel, Field, validator, constr
from typing import Optional
from uuid import UUID, uuid4

from whyqd.models import CitationModel, DataSourceModel, CrosswalkModel, VersionModel


class TransformModel(BaseModel):
    uuid: UUID = Field(default_factory=uuid4, description="Automatically generated unique identity for the method.")
    name: constr(strip_whitespace=True, to_lower=True) = Field(
        ...,
        description="Machine-readable term to uniquely address this method. Cannot have spaces. CamelCase or snake_case.",
    )
    title: Optional[str] = Field(None, description="A human-readable version of the method name.")
    description: Optional[str] = Field(
        None,
        description="A complete description of the method. Depending on how complex your work becomes, try and be as helpful as possible to 'future-you'. You'll thank yourself later.",
    )
    dataSource: Optional[DataSourceModel] = Field(
        None,
        description="Tabular source data defined by the crosswalk source schema to produce destination data conforming to a destination schema.",
    )
    dataDestination: Optional[DataSourceModel] = Field(
        None,
        description="Tabular source data defined by the crosswalk destination schema and resulting from execution of the crosswalk.",
    )
    crosswalk: Optional[CrosswalkModel] = Field(
        None,
        description="A method defining the source schema to destination schema crosswalk to be applied to tabular source data.",
    )
    citation: Optional[CitationModel] = Field(None, description="Optional full citation for the transform.")
    version: list[VersionModel] = Field(default=[], description="Version and update history for the method.")

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True

    @validator("name")
    def name_space(cls, v):
        return "_".join(v.split(" ")).lower()



"""
Actions
-------

Definitions and validation for actions anchoring structures in methods. Describes how an Action performs its roll. 
New transforms inherit from the core.BaseSchemaAction class and are automatically available as a plugin.

Key functions are:

  - Define action settings
  - Specify any modifiers
  - Validate action structure
  - Perform action dataframe transformation

The structure of an action (JSON or dict) saved in a method (and presented to the `action.transform` funtion)::

    [
        {
            "description": "",
            "name": "ACTION",
            "structure": [
                "field"
            ],
            "title": "Title",
            "type": "action"
        },
        {
            "name": "field",
            "type": "type"
        }
    ]

Where the first term in the list if the action (as would appear in `action.settings`) and the subsequent fields
must conform to the structure defined in the action's structure, but with a format that can be accessed in the 
`action.transform` function. Normally this would be a `name` and `type`, but any created value would have a `value`
instead of a `name`.
"""
import pkgutil
import importlib.util as _importlib_util
from pathlib import Path

__action_path__ = str(Path(__file__).resolve().parent)

def load_dynamic(name, module_path):
    # https://bugs.python.org/issue43540
    module_path = f"{module_path}/{name}"
    if not module_path.endswith(".py"):
        module_path = f"{module_path}.py"
    spec = _importlib_util.spec_from_file_location(name, module_path)
    module = _importlib_util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.Action

actions = {
    name.upper(): load_dynamic(name, finder.path)
    for finder, name, _ispkg in pkgutil.iter_modules([__action_path__])
}
default_actions = [a for a in [actn().settings for actn in actions.values()]]



from __future__ import annotations
from pydantic import Json, BaseModel
from collections.abc import MutableMapping
from pathlib import Path
from datetime import date, datetime
import json

from whyqd.parsers import CoreParser
from whyqd.models import VersionModel, CitationModel


class BaseDefinition:
    """Core shared base definition functionality."""

    def __init__(self, *, model_name: str) -> None:
        self.core = CoreParser()
        self.model = None
        self.model_name = model_name  # method / schema / transform / etc.

    def __repr__(self) -> str:
        """Returns the string representation of the model."""
        if self.model:
            return f"{self.model_name.title()}: `{self.model.name}`"
        return f"{self.model_name.title()}"

    @property
    def describe(self) -> dict[str, None] | None:
        """Get the model name, title and description.

         - name: Term used for filename and referencing.
         - title: Human-readable term used as name.
         - description: Detailed description for the model. Reference its objective and use-case.

        Returns:
          A dictionary with the `name`, `title` and `description` for the `Definition`.
        """
        if self.model:
            return {
                "name": self.model.name,
                "title": self.model.title,
                "description": self.model.description,
            }
        return None

    def _refresh_model_terms(self) -> None:
        """Any process to be run to ensure concord between this definition and its model."""
        pass

    def _refresh_model_fields(self) -> None:
        """Refreshes the list of model fields and connects them to the Field CRUD"""
        pass

    #########################################################################################
    # MANAGE CITATION
    #########################################################################################

    def get_citation(self) -> dict[str, str | dict[str, str]]:
        """Get the citation as a dictionary.

        Raises:
          ValueError: If no citation has been declared or the build is incomplete.

        Returns:
          A dictionary conforming to the CitationModel.
        """
        if not self.model.citation:
            raise ValueError("No citation has been declared yet.")
        return self.model.citation.dict(by_alias=True, exclude_defaults=True, exclude_none=True)

    def set_citation(self, *, citation: CitationModel) -> None:
        """Update or create the citation.

        **whyqd** is designed to support a research process and ensure citation of the incredible work done by
        research-based data scientists.

        A citation has the following definable fields::

        - **author**: The name(s) of the author(s) (in the case of more than one author, separated by `and`),
        - **title**: The title of the work,
        - **url**: The URL field is used to store the URL of a web page or FTP download. It is a non-standard BibTeX field,
        - **publisher**: The publisher's name,
        - **institution**: The institution that was involved in the publishing, but not necessarily the publisher,
        - **doi**: The doi field is used to store the digital object identifier (DOI) of a journal article, conference paper,
          book chapter or book. It is a non-standard BibTeX field. It's recommended to simply use the DOI, and not a DOI link,
        - **month**: The month of publication (or, if unpublished, the month of creation). Use three-letter abbreviation,
        - **year**: The year of publication (or, if unpublished, the year of creation),
        - **licence**: The terms under which the associated resource are licenced for reuse,
        - **note**: Miscellaneous extra information.

        Parameters:
          citation: A dictionary conforming to the CitationModel.

        Example:
          ```python
          import whyqd as qd

          datasource = qd.DataSourceDefinition()
          datasource.derive_model(source=SOURCE_DATA, mimetype=MIMETYPE)
          citation = {
              "author": "Gavin Chait",
              "month": "feb",
              "year": 2020,
              "title": "Portsmouth City Council normalised database of commercial ratepayers",
              "url": "https://github.com/whythawk/whyqd/tree/master/tests/data",
              "licence": "Attribution 4.0 International (CC BY 4.0)",
          }
          datasource.set_citation(citation=citation)
          ```
        """
        # Create a temporary CitationModel
        self.model.citation = self.core.create_or_update_model(
            modelType=CitationModel, source=citation, model=self.model.citation
        )

    #########################################################################################
    # SAVE UTILITIES
    #########################################################################################

    def exclude_uuid(self, *, model: BaseModel | dict):
        # https://stackoverflow.com/a/49723101/295606
        if isinstance(model, BaseModel):
            model = model.dict(by_alias=True, exclude_defaults=True, exclude_none=True, exclude={"uuid": True})
        excluded = {}
        for key, field in model.items():
            if key != "uuid":
                if isinstance(field, MutableMapping):
                    excluded[key] = self.exclude_uuid(model=field)
                elif field and isinstance(field, list) and isinstance(field[0], MutableMapping):
                    # Assuming field consistency ... no mixed Mutables and non-Mutables
                    excluded[key] = []
                    for f in field:
                        excluded[key].append(self.exclude_uuid(model=f))
                else:
                    if isinstance(field, (datetime, date)):
                        # To make sure json.dumps doesn't run into datetime errors
                        field = field.isoformat()
                    excluded[key] = field
        return excluded

    def get_json(self, hide_uuid: bool = False) -> Json | None:
        """Get the json model.

        Parameters:
          hide_uuid: Hide all UUIDs in the nested JSON output. Mostly useful for validation assertions where the only
                     differences between sources are the UUIDs.

        Returns:
          Json-conforming output, or None.
        """
        self._refresh_model_terms()
        if self.model and not hide_uuid:
            return self.model.json(by_alias=True, exclude_defaults=True, exclude_none=True)
        elif self.model and hide_uuid:
            return json.dumps(self.exclude_uuid(model=self.model))
        return None

    def save(
        self,
        directory: str | None = None,
        filename: str | None = None,
        created_by: str | None = None,
        hide_uuid: bool = False,
    ) -> bool:
        """Save model as a json file.

        Parameters:
          directory:  Defaults to working directory
          filename:  Defaults to model name
          created_by:  Declare the model creator/updater
          hide_uuid:  Hide all UUIDs in the nested JSON output.

        Returns:
          Boolean True if saved.
        """
        self._refresh_model_terms()
        if not self.model:
            raise ValueError(f"{self.model_name.title()} model does not exist.")
        if not directory:
            directory = self.core.default_directory
        else:
            directory = self.core.check_path(directory=directory)
        if not filename:
            filename = f"{self.model.name}.{self.model_name}"
        if filename.split(".")[-1] != self.model_name:
            filename += f".{self.model_name}"
        if isinstance(directory, str):
            directory = Path(directory)
        path = directory / filename
        if "version" in self.model.dict():
            update = VersionModel(**{"description": f"Save {self.model_name}."})
            if created_by:
                update.name = created_by
            self.model.version.append(update)
        return self.core.save_file(data=self.get_json(hide_uuid=hide_uuid), source=path)



from __future__ import annotations
from typing import Optional, List, Dict, Any
from pydantic import BaseModel, Field, validator, constr
from uuid import UUID, uuid4
import string

from whyqd.dtypes import MimeType
from whyqd.models import CitationModel, ColumnModel


class DataSourceAttributeModel(BaseModel):
    # https://docs.pydantic.dev/latest/usage/models/#custom-root-types
    __root__: Dict[str, Any]

    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]

    def __setitem__(self, item, term):
        self.__root__[item] = term

    @property
    def terms(self):
        return self.__root__


class DataSourceModel(BaseModel):
    uuid: UUID = Field(
        default_factory=uuid4, description="Automatically generated unique identity for the data source."
    )
    title: Optional[str] = Field(None, description="A descriptive name for the data source.")
    description: Optional[str] = Field(
        None,
        description="A complete description of the data source. Depending on how complex your work becomes, try and be as helpful as possible to 'future-you'. You'll thank yourself later.",
    )
    path: constr(strip_whitespace=True) = Field(..., description="Full path to valid source data file.")
    mime: MimeType = Field(..., description="Mime type for source data. Automatically generated.")
    header: Optional[int] = Field(
        0, description="Row (0-indexed) to use for the column labels of the parsed DataFrame. "
    )
    sheet_name: Optional[str] = Field(
        None,
        description="Needed if MimeType is either of XLS or XLSX, and the data consists of multiple sheets.",
    )
    names: Optional[List[str]] = Field(
        [],
        description="If the source data has no header row, explicitly pass a list of column names - in the correct order - to address the data. May not be known until after initial wrangling.",
    )
    columns: List[ColumnModel] = Field(
        ..., description="List of ColumnModel with names & type derived from the source data. In table order."
    )
    preserve: Optional[List[str]] = Field(
        [],
        description="List of columns where variable type guessing must be prevented and the original data preserved. Critical for foreign key references with weird formats, like integers with leading `0`.",
    )
    key: Optional[List[str]] = Field(
        [],
        description="A list of key columns required before merging of multiple data sources into an interim data source.",
    )
    attributes: Optional[DataSourceAttributeModel] = Field(
        {},
        description="Optional open dictionary for reader-specific attributes for Pandas' API. E.g. 'quoting' for the CSV library.",
    )
    checksum: str = Field(default=..., description="Checksum for source data file. Based on Blake2b.")
    index: Optional[int] = Field(None, description="Count of rows in source data table.")
    citation: Optional[CitationModel] = Field(None, description="Optional full citation for the source data.")

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True

    @validator("preserve")
    def preserve_check(cls, v, values, **kwargs):
        if not values["columns"] or set(v).difference([c.name for c in values["columns"]]):
            raise ValueError(
                f"Columns ({set(v).difference([c.name for c in values['columns']])}) not in source data columns."
            )
        return v

    @validator("key")
    def key_check(cls, v, values, **kwargs):
        if not values["columns"] or set(v).difference([c.name for c in values["columns"]]):
            raise ValueError(
                f"Key columns ({set(v).difference([c.name for c in values['columns']])}) not in source data columns."
            )
        return v

    @property
    def hexed_filename(self) -> str:
        return f"{str(self.uuid)}.{MimeType(self.mime).name.lower()}"

    @property
    def name(self) -> str:
        root = "".join(c for c in self.path.split("/")[-1] if c in f"-_. {string.ascii_letters}{string.digits}")
        if self.sheet_name:
            return f"{root}::{self.sheet_name}"
        return root



from __future__ import annotations
from pydantic import BaseModel, Field, validator


class ModifierModel(BaseModel):
    name: str = Field(..., description="Single-character modifier.")
    title: str = Field(..., description="Human-readable title for the name. Acts as a description.")

    class Config:
        validate_assignment = True
        use_enum_values = True
        # anystr_strip_whitespace = True

    @validator("name")
    def check_name(cls, v):
        if len(v) != 1:
            raise ValueError(f"Modifier provided ({v}) must be a single character.")
        return v



from __future__ import annotations
from typing import List, Optional
from pydantic import BaseModel, Field, validator


class CategoryActionModel(BaseModel):
    """Category Model - generated from the Category module. A type of BaseCategoryAction."""

    name: str = Field(..., description="Name of the Action. Uppercase.")
    title: str = Field(..., description="Title of the Action. Regular case.")
    description: str = Field(..., description="Description of the purpose for performing this action.")
    structure: Optional[List[str]] = Field(
        default=[],
        description="The structure of an action depends on either 'boolean' or 'unique' action terms.",
    )

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True

    @validator("structure")
    def check_valid_models(cls, v):
        for s in v:
            if not (s in ["boolean", "unique"]):
                raise ValueError(f"Structure ({s}) must be of either `boolean` or `unique`.")
        return v



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseMorphAction

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseMorphAction):
    """Remove all duplicated rows from a DataFrame.

    !!! tip "Script template"
        ```python
        "DEDUPE"
        ```
    """

    def __init__(self) -> None:
        self.name = "DEDUPE"
        self.title = "Deduplicate"
        self.description = "Remove all duplicated rows from a DataFrame."
        self.structure = []

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: None = None,
        source: None = None,
        rows: None = None,
        source_param: None = None,
        destination_param: None = None,
    ) -> pd.DataFrame:
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.drop_duplicates.html
        return df.drop_duplicates()



from __future__ import annotations
from typing import List, TYPE_CHECKING

from uuid import uuid4

from whyqd.parsers import CoreParser
from whyqd.models import MorphActionModel

if TYPE_CHECKING:
    import modin.pandas as pd
    from whyqd.models import ColumnModel


class BaseMorphAction:
    """Morphs differ from Actions in that they normally act to reshape entire tables rather than manipulate columns.

    * Morph actions are not permitted to be nested, i.e. they are stand-alone ActionScripts.
    * May result in changes to column or row references that must be accounted for in other actions.

    Morphs inherit from this base class which describes the core functions and methodology for a Morph. They
    should redefine `name`, `title`, `description`, and `structure`, as well as produce a `transform` function.
    Everything else will probably remain as defined, but particularly complex Morphs should modify as required.

    `structure` can be an empty list, but a morph may be defined by these parameters:

    * `rows`: the specific rows effected by the morph, a `list` of `int`
    * `columns`: the specific columns effected by the morph, a `list` of `ColumnModel` or `FieldModel`.

    A standard script is::

        "ACTION > [columns] < [rows]"

    Where:

    * the presence and order of the arrays is set by `structure`,
    * `columns` are indicated by `>`, and
    * `rows` are indicated by `<`.
    """

    def __init__(self) -> None:
        self.core = CoreParser()
        self.name = ""
        self.title = ""
        self.description = ""
        # `structure` defines the format in which parameters for a morph are written, and validated in
        # `validate` can be any of `rows` or `columns`.
        # Where new columns are created, these will be randomly-generated and can be renamed as required.
        self.structure = []

    @property
    def settings(self) -> MorphActionModel:
        """
        Returns the dict representation of the Morph.

        Raises
        ------
        NameError if parameters don't yet exist.

        Returns
        -------
        dict
            Dict representation of a Morph.
        """
        morph_settings = {
            "name": self.name,
            "title": self.title,
            "type": "morph",
            "description": self.description,
            "structure": self.structure,
        }
        return MorphActionModel(**morph_settings)

    def transform(self, *, df: pd.DataFrame, rows: List[int], columns: List[ColumnModel]) -> pd.DataFrame:
        """
        Perform a transformation. This function must be overridden by child Morphs and describe a unique
        new method.

        .. warning:: Assumes `validates` has been run.

        Parameters
        ----------
        df: DataFrame
            Working data to be transformed
        **kwargs:
            Other fields which may be required in custom transforms

        Returns
        -------
        Dataframe
            Containing the implementation of the Morph
        """
        return df

    def _column_renames_to_index(self, *, df: pd.DataFrame, columns: str, hex: str) -> pd.DataFrame:
        """
        Rename randomly-generated column labels with their predictable index equivalent.

        Parameters
        ----------
        df: DataFrame
        columns: List[str]
        hex: str
            The random string used in the generator.

        Returns
        -------
        DataFrame
        """
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rename.html
        # {'from': 'too'}
        renames = {c: c.replace(hex, f"idx_{df.columns.get_loc(c)}") for c in columns}
        return df.rename(index=str, columns=renames)

    def _generate_hex(self):
        return str(uuid4().hex)[:4]



from __future__ import annotations

from whyqd.crosswalk.base import BaseCategoryAction


class Action(BaseCategoryAction):
    """Associate category terms in a source data field to a specific categorical term in a destination field. If there
     are numerous source and destination terms, then individual scripts are needed for each.

    !!! tip "Script template"
         ```python
         "CATEGORISE > 'destination_field'::'destination_term' < 'source_field'::['source_term']"
         ```

         - If `destination_term` is a boolean (`True` or `False`) then `source_term` hits will assign that boolean value.
         - If `source_term` is a boolean (`True` or `False`) then any value in that column will be assigned `True` or
           `False` and convey the `destination_term` value.
           If the `source_term` is boolean, then there **must** only be one source term.

    !!! example
        ```python
        "CATEGORISE > 'occupation_state_reliefs'::'other' < 'Current Relief Type'::['Sports Club (Registered CASC)', 'Mandatory']"
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "CATEGORISE"
        self.title = "Categorise"
        self.description = "Associate category terms in a source data field to a specific categorical term in a destination field. If there are numerous source and destination terms, then individual scripts are needed for each."
        self.structure = ["boolean", "unique"]



from __future__ import annotations
from typing import TYPE_CHECKING
import modin.pandas as pd
import numpy as np

from whyqd.parsers import DataSourceParser, ScriptParser

if TYPE_CHECKING:
    from whyqd.models import CategoryActionModel, CategoryModel, FieldModel


class BaseCategoryAction:
    """Category Actions are support utilities for CATEGORY actions. These inherit from this base class which describes
    the core functions and methodology for this support Action.

    Actions should redefine `name`, `title`, `description`, and `structure`.

    `structure` is defined by these parameters:

    * `boolean`: terms will be unique.
    * `unique`: the specific columns effected by the morph, a `list` of `ColumnModel` or, rarely, `FieldModel`.

    A standard script is::

        "ACTION > 'destination_column::destination_category' < ['unique_term', 'unique_term', etc.]"

    Where the structure of the source array is defined by the ACTION.
    """

    def __init__(self) -> None:
        self.reader = DataSourceParser()
        self.parser = ScriptParser()
        self.name = ""
        self.title = ""
        self.description = ""
        self.structure = []

    @property
    def settings(self) -> CategoryActionModel:
        """
        Returns the CategoryActionModel representation of the Action.

        Returns
        -------
        CategoryActionModel
            CategoryActionModel representation of an Action.
        """
        from whyqd.models import CategoryActionModel

        action_settings = {
            "name": self.name,
            "title": self.title,
            "description": self.description,
            "structure": self.structure,
        }
        return CategoryActionModel(**action_settings)

    def parse(self, *, script: str) -> dict[str, str]:
        """Base parser for the CategoryAction script. Produces required terms and validates against this
        CategoryAction's structure requirements.

        Script is of the form::

            "ACTION > 'destination_column'::term < 'source_column'::[term]"

        Which is inherited as::

            {
                "action": ACTION,
                "destination": 'destination_column',
                "category": term,
                "source": 'source_column',
                "source_category": [term]
            }

        Parameters
        ----------
        script: str
            An action script.

        Raises
        ------
        ValueError for any parsing errors.

        Returns
        -------
        dict
            Parsed dictionary of validated split strings for further processing.
        """
        parsed = {}
        # Get unique assignment terms
        root = self.parser.get_split_terms(script=script, by="<")
        if len(root) != 2:
            raise ValueError(f"Category action script has no source terms ({script}).")
        # Extract source terms
        source_terms = self.parser.get_split_terms(script=root[1], by="::")
        parsed["source"] = source_terms[0]
        if len(source_terms) == 2:
            parsed["source_category"] = source_terms[1]
        root = self.parser.get_split_terms(script=root[0], by=">")
        # Extract destination terms
        if len(root) != 2:
            raise ValueError(f"Category action script has no destination terms ({script}).")
        destination_terms = self.parser.get_split_terms(script=root[1], by="::")
        if len(destination_terms) != 2:
            raise ValueError(f"Category action script has no destination category term ({script}).")
        parsed["destination"] = destination_terms[0]
        parsed["category"] = destination_terms[1]
        # And the action term
        parsed["action"] = root[0]
        return parsed

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: FieldModel,
        category: CategoryModel = None,
        assigned: list[CategoryModel] | list[int] | None = None,
    ) -> pd.DataFrame:
        """
        Produce categories from terms or headers. There are three categorisation options:

            1. Term-data are categories derived from values in the data,
            2. Header-data are terms derived from the header name and boolean True for any value,
            3. "Boolean"-data, the category itself is True/False. Default in a boolean is True.

        .. note:: Categorisation is a special case, requiring both method fields, and method categories, which it can do
        only if the `destination` is a schema field, which has the required term definitions.

        Parameters
        ----------
        df: DataFrame
            Working data to be transformed
        destination: FieldModel
            Destination FieldModel for the result of the Action.
        source: list of ColumnModel and / or ModifierModel
            List of source columns and modifiers for the action.
        assigned: list of dict or list of int
            Each dict has values for: Assignment ACTION, destination schema field, schema category, source data column,
            and a list of source data column category terms assigned to that schema category.
            Can also assign specific row values to a category for pivots.

        Returns
        -------
        Dataframe
            Containing the implementation of the Action
        """
        # Defaults
        default = None
        if destination.constraints and destination.constraints.default:
            default = destination.constraints.default.name
        if destination.name in df.columns and destination.dtype != "array":
            default = df[destination.name]
        # Conditions
        conditions = df[source.name]
        if len(assigned) == 1 and isinstance(assigned[0].name, bool):
            # Conditions are contingent on values in a column
            # If assigned is `True`, then values are `True`, else values are `False` and nulls are `True`
            if source.dtype and source.dtype not in ["string", "object"]:
                conditions = self.reader.coerce_column_to_dtype(column=df[source.name], coerce=source.dtype)
                if conditions.dtype in ["float64", "int64", "Float64", "Int64"]:
                    conditions = conditions.replace({0: np.nan})
                if conditions.dtype in ["datetime64[ns]"]:
                    conditions = conditions.apply(self.reader.parse_dates_coerced)
            if assigned[0].name:
                # i.e. values are assigned True
                conditions = pd.Series([not (pd.isnull(x) or not x) for x in conditions.tolist()])
            else:
                # i.e. values are assigned False
                conditions = pd.Series([(pd.isnull(x) or not x) for x in conditions.tolist()])
        else:
            # The conditional column values are categorised directly
            conditions = conditions.isin([c.name for c in assigned])
        # Array-based categorisation adds additional terms to existing columns
        if destination.dtype == "array":
            new_column = []
            if destination.name in df.columns:
                new_column = [df[destination.name].tolist()]
            if destination.dtype == "array" and not conditions.empty:
                new_column.append(
                    [
                        [x] if not isinstance(x, list) else x
                        for x in np.where(conditions, category.name, default).tolist()
                    ]
                )
            # Sorted to avoid hashing errors later ...
            if len(new_column) > 1:
                # Moving sorting to the hashing function so we can maintain None's for ordered lists
                # Permits reconstruction of datasets using array transformations
                # https://stackoverflow.com/a/18411610
                # new_column = [sorted(list(set([c for clist in x for c in clist])), key=lambda x: (x is None, x)) for x in zip(*new_column)]
                new_column = [[c for clist in x for c in clist] for x in zip(*new_column)]
            else:
                new_column = [[x] if not isinstance(x, list) else x for x in new_column[0]]
            # for n in new_column:
            #     if None in n:
            #         n.remove(None)
            df[destination.name] = new_column
        else:
            df[destination.name] = np.where(conditions, category.name, default)
        return df



from __future__ import annotations
from uuid import UUID
import modin.pandas as pd

from whyqd.crud.base import CRUDBase
from whyqd.models import FieldModel, ConstraintsModel, CategoryModel
from whyqd.dtypes import FieldType


class CRUDField(CRUDBase[FieldModel]):
    """Create, Read, Update and Delete Field Models. Usually instantiated as part of a
    [SchemaDefinition](schema.md) and accessed as `.fields`.

    [Base CRUD operations](basecrud.md) are common for both `CRUDField` and `CRUDAction`.

    Example:
      ```python
      import whyqd as qd

      schema: qd.models.SchemaModel = {
          "name": "urban_population",
          "title": "Urban population",
          "description": "Urban population refers to people living in urban areas as defined by national statistical offices.",
      }
      fields: list[qd.models.FieldModel] = [
          {
              "name": "indicator_code",
              "title": "Indicator Code",
              "type": "string",
              "description": "World Bank code reference for Indicator Name.",
              "constraints": {"required": True},
          },
          {
              "name": "country_name",
              "title": "Country Name",
              "type": "string",
              "description": "Official country names.",
              "constraints": {"required": True},
          },
          {
              "name": "country_code",
              "title": "Country Code",
              "type": "string",
              "description": "UN ISO 3-letter country code.",
              "constraints": {"required": True},
          },
          {
              "name": "indicator_name",
              "title": "Indicator Name",
              "type": "string",
              "description": "Indicator described in the data series.",
              "constraints": {"required": True},
          },
          {
              "name": "year",
              "title": "Year",
              "type": "year",
              "description": "Year of release.",
              "constraints": {"required": True},
          },
          {
              "name": "values",
              "title": "Values",
              "type": "number",
              "description": "Value for the Year and Indicator Name.",
              "constraints": {"required": True},
          },
      ]
      schema_destination = qd.SchemaDefinition()
      schema_destination.set(schema=schema)
      schema_destination.fields.add_multi(terms=fields)
      ```

    Parameters:
      model: The Pydantic model on which CRUD operations are based.
    """

    def get_required(self) -> list[FieldModel]:
        """Return a list of all required fields.

        Returns:
          A list of required fields.
        """
        return [f for f in self.multi if f.constraints and f.constraints.required]

    def get_constraints(self, *, name: str | UUID) -> ConstraintsModel | None:
        """Get the constraint parameters for a specific field defined in this schema, called by a unique
        `name` already in the schema.

        Parameters:
          name: Specific name or reference UUID for a field already in the Schema.

        Returns:
          A ConstraintsModel or None.
        """
        field = self.get(name=name)
        if not field:
            raise ValueError(f"FieldModel {name} does not exist in the schema.")
        return field.constraints

    def set_constraints(self, *, name: str | UUID, constraints: ConstraintsModel | None) -> None:
        """Set the constraint parameters for a specific field to define this schema, called by a unique
        `name` already in the schema.

        Parameters:
          name: Specific name or reference UUID for a field already in the Schema
          constraints: A dictionary conforming to the ConstraintsModel, or None. If None, then constraints are deleted.
        """
        old_constraints = self.get_constraints(name=name)
        if not constraints:
            old_constraints = None
        else:
            new_constraints = ConstraintsModel(**constraints)
            if old_constraints:
                old_constraints = old_constraints.copy(update=new_constraints.dict(exclude_unset=True))
            else:
                old_constraints = new_constraints
        self.get(name=name).constraints = old_constraints

    def set_categories(
        self,
        *,
        name: str,
        terms: pd.DataFrame | pd.Series | list[str] | None = None,
        as_bool: bool = False,
        has_array: bool = False,
    ) -> None:
        """Derive unique category constraints from a data column for a specific field, called by a unique `name`
        already in the schema.

        Parameters:
          name:  Specific name for a field already in the Schema
          terms:  Data provided as a table or column containing string terms, as arrays or individual terms
          as_bool:  Set the category terms as booleans, `True` and `False`
          has_array:  Set `True` if data terms are arrays, or the Field is of type `FieldType.ARRAY`
        """
        field = self.get(name=name)
        if not field.constraints:
            field.constraints = ConstraintsModel()
        if as_bool:
            field.constraints.category = [CategoryModel(**{"name": term}) for term in [True, False]]
        else:
            if not isinstance(terms, list):
                terms = terms[name].unique()
            if has_array or field.dtype == FieldType.ARRAY:
                # Multiple categories in a row
                field.dtype = FieldType.ARRAY
                # This will only work where it's a 2D array, which it 'should' be
                # https://stackoverflow.com/a/38900498/295606
                terms = [
                    x for inner in [[item] if not isinstance(item, list) else item for item in terms] for x in inner
                ]
            field.constraints.category = [
                CategoryModel(**{"name": term})
                for term in terms
                if not (term is pd.NA or pd.isnull(term) or term is None)
            ]

    def get_category(self, *, name: str, category: bool | str) -> CategoryModel | None:
        """Get a specific field from the list of fields defining this schema, called by a unique `name`.

        Parameters:
          name:  Field names must be unique, so a valid `name` in the field list will have no collisions.
          category:  Category name for a destination category term from the schema.

        Raises:
          ValueError: If `FieldModel` has not `category` in `ConstraintModel`.

        Returns:
          A list of CategoryModel, or None of none are defined.
        """
        if self.multi:
            field_categories = self.get_constraints(name=name)
            if not field_categories and isinstance(category, bool):
                # Bools have default True, False categories
                # Create them now if they don't exist
                # TODO: put this somewhere more useful, and where the user can set default is True/False
                true = CategoryModel(**{"name": True})
                false = CategoryModel(**{"name": False})
                fld = self.get(name=name)
                fld.constraints = ConstraintsModel(**{"default": true, "enum": [true, false]})
            field_categories = self.get_constraints(name=name).category
            if not field_categories:
                raise ValueError(f"Field ({name}) has no `category` constraints.")
            # https://stackoverflow.com/a/31988734/295606
            return next((f for f in field_categories if f.name == category or f.uuid.hex == category), None)
        return None



from __future__ import annotations
from typing import TYPE_CHECKING
from pydantic import Json
import json
import sys
import hashlib
from urllib.parse import urlparse
import urllib
import posixpath
from uuid import uuid4
from datetime import date, datetime
from pathlib import Path, PurePath
import modin.pandas as pd
import numpy as np

# from pandas.util import hash_pandas_object
import locale
import randomname
import warnings

from whyqd.config.settings import settings

if TYPE_CHECKING:
    from whyqd.models import SchemaModel, CrosswalkModel

try:
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
except locale.Error:
    # Readthedocs has a problem, but difficult to replicate
    locale.setlocale(locale.LC_ALL, "")


class CoreParser:
    """Core functions for file and path management, and general ad-hoc utilities."""

    def __init__(self) -> None:
        self.default_directory = settings.WHYQD_DIRECTORY

    ###################################################################################################
    ### AD-HOC UTILITIES
    ###################################################################################################

    def get_now(self):
        return date.isoformat(datetime.now())

    def chunks(self, *, lst: list, n: int) -> list:
        """Yield successive n-sized chunks from l."""
        # https://stackoverflow.com/a/976918/295606
        for i in range(0, len(lst), n):
            yield lst[i : i + n]

    def show_warning(self, message: str) -> None:
        warnings.warn(message, UserWarning)

    ###################################################################################################
    ### MODEL UTILITIES
    ###################################################################################################

    def create_or_update_model(
        self,
        *,
        modelType: type[SchemaModel] | type[CrosswalkModel],
        source: Path | str | SchemaModel | CrosswalkModel | None = None,
        model: SchemaModel | CrosswalkModel | None = None,
    ) -> SchemaModel | CrosswalkModel:
        """Update or create a model. Must have a default `name` field as the only dependency.

        Parameters
        ----------
        modelType: type of SchemaModel, CrosswalkModel
            The Pydantic function for the model type.
        source: Path, str or modelType, default None
            Any approach to loading a source.
        model: modelType, default None
            An existing model to be updated.
        """
        # Create a temporary model
        if not source:
            updated_model = modelType(**{"name": randomname.get_name()})
        else:
            if isinstance(source, (Path, str)):
                updated_model = modelType(**self.load_json(source=source))
            elif isinstance(source, (Json, dict)):
                updated_model = modelType(**source)
            elif isinstance(source, modelType):
                updated_model = source
            else:
                raise ValueError(
                    "`source` must either be a path to a source file, or a dictionary conforming to the model type."
                )
        # And update the original data
        # https://fastapi.tiangolo.com/tutorial/body-updates/#partial-updates-with-patch
        # Which doesn't really work properly as it returns dictionaries ... validation / init doesn't happen
        # This approach may be less efficient, but it does actually parse
        if model:
            model = model.dict(by_alias=True, exclude_defaults=True, exclude_none=True)
            model |= updated_model.dict(exclude_unset=True)
            return modelType(**model)
        else:
            return updated_model

    ###################################################################################################
    ### PATH MANAGEMENT
    ###################################################################################################

    def check_path(self, *, directory: str) -> Path:
        if isinstance(directory, str):
            directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)
        return directory

    def check_source(self, *, source: str) -> bool:
        if isinstance(source, str):
            source = Path(source)
        return source.exists()

    def get_path(self) -> PurePath:
        # whyqd/core/common ... .parent.parent -> whyqd/
        return Path(__file__).resolve().parent.parent

    def check_uri(self, *, source: str) -> bool:
        # https://stackoverflow.com/a/38020041
        try:
            result = urlparse(source)
            return all([result.scheme, result.netloc, result.path])
        except (ValueError, AttributeError):
            return False

    def check_path_or_uri(self, *, source: str) -> bool:
        if not isinstance(source, str):
            source = str(source)
        return any([Path(source).exists(), self.check_uri(source=source)])

    def rename_file(self, *, source: str, newname: str, add_suffix: bool = False) -> str:
        p = Path(source)
        if add_suffix:
            newname = f"{newname}.{p.suffix}"
        p.rename(Path(p.parent, newname))
        return newname

    def delete_file(self, *, source: Path | str):
        if isinstance(source, str):
            source = Path(source)
        try:
            assert sys.version_info >= (3, 8)
        except AssertionError:
            source.unlink()
        else:
            source.unlink(missing_ok=True)

    def download_uri_source(self, source: str, directory: Path | str | None = None) -> Path:
        """Downloads a source at a remote uri, and returns a Path for that downloaded source.

        Parameters
        ----------
        source: str
            URI path to source.
        directory: Path | str | None

        Returns
        -------
        Path to local source.
        """
        request = urllib.request.Request(source, method="HEAD")
        request = urllib.request.urlopen(request).info()
        filename = request.get_filename()
        if not filename:
            #  https://stackoverflow.com/a/11783319/295606
            source_path = urllib.request.urlsplit(source).path
            filename = posixpath.basename(source_path)
        if not filename:
            # Make something up ...
            filename = f"temporary-{uuid4()}"
        if not directory:
            directory = self.default_directory
        local_source = Path(directory) / filename
        if not self.check_source(source=local_source):
            urllib.request.urlretrieve(source, local_source)
        return local_source

    ###################################################################################################
    ### CHECKSUMS
    ###################################################################################################

    def get_checksum(self, *, source: str) -> str:
        # https://stackoverflow.com/a/47800021
        checksum = hashlib.blake2b()
        with open(source, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                checksum.update(chunk)
        return checksum.hexdigest()

    def get_data_checksum(self, *, df: pd.DataFrame) -> str:
        # The destination data does not have a valid checksum for the file itself, only the data.
        # df_checksum = hash_pandas_object(df.astype("string")._to_pandas(), index=True).values
        # Need to ensure arrays are all of the same type. Easiest is to coerce np arrays to lists
        for column in df.columns:
            if isinstance(df.iloc[0][column], np.ndarray):
                df[column] = df[column].apply(lambda x: x.tolist())
        df_string = df.astype("string").astype("string").to_string(index=False).encode("utf-8")
        checksum = hashlib.blake2b
        return checksum(df_string).hexdigest()

    ###################################################################################################
    ### JSON & FILE LOAD & SAVE
    ###################################################################################################

    def load_json(self, *, source: str) -> dict:
        """
        Load and return a JSON file, if it exists.

        Paramaters
        ----------
        source: the filename & path to open

        Raises
        ------
        JSONDecoderError if not a valid json file
        FileNotFoundError if not a valid source

        Returns
        -------
        dict
        """
        with open(source, "r") as f:
            try:
                return json.load(f)
            except json.decoder.JSONDecodeError:
                e = "File at `{}` not valid json.".format(source)
                raise json.decoder.JSONDecodeError(e)

    def save_json(self, *, data: dict, source: str) -> bool:
        """
        Save a dictionary as a json file.

        Parameters
        ----------
        data: dictionary to be saved
        source: the filename to save, including path

        Returns
        -------
        bool, True if saved.
        """
        with open(source, "w") as f:
            json.dump(data, f, indent=4, sort_keys=True, default=str)
        return True

    def save_file(self, *, data: json, source: str) -> bool:
        """Save json to file.

        Parameters
        ----------
        data: dictionary to be saved
        source: the filename to save, including path

        Returns
        -------
        bool, True if saved.
        """
        with open(source, "w") as f:
            f.write(data)
        return True



from pydantic import BaseSettings, Field
from typing import Optional


class Settings(BaseSettings):
    # 10**9 == 1gb ==> try 8gb?
    # dev_shm must be aligned in docker-compose, e.g.
    #   shm_size: "8gb"
    WHYQD_MEMORY: Optional[int] = Field(default=None)
    WHYQD_CPUS: Optional[int] = Field(default=None)
    WHYQD_SPILLWAY: Optional[str] = Field(default="/spill")
    WHYQD_DIRECTORY: Optional[str] = Field(default="")
    WHYQD_DEFAULT_MIMETYPE: Optional[str] = Field(default="application/vnd.apache.parquet")

    class Config:
        case_sensitive = True
        env_file = ".env"
        env_file_encoding = "utf-8"


settings = Settings()



from __future__ import annotations
from typing import TYPE_CHECKING
import modin.pandas as pd
import numpy as np

from whyqd.crosswalk.base import BaseCategoryAction

if TYPE_CHECKING:
    from whyqd.models import FieldModel


class Action(BaseCategoryAction):
    """Convert row-level categories into field categorisations.

    !!! tip "Script template"
        ```python
        "PIVOT_CATEGORIES > 'destination_field' < 'source_field'::[int, int, int, etc.]"
        ```

        Where `int` contains the rows that define the categories, and `field` are the fields to include. Makes several
        assumptions:

        - Rows may contain more than one category
        - All terms in indexed rows in the same field are related
        - Categories are assigned downwards to all rows between indices
        - The last indexed category is assigned to all rows to the end of the table

    !!! example
        ```python
        "PIVOT_CATEGORIES > 'NewField' < 'Field'::[1, 5]"
        ```

        Will transform:

        | ID | Field |
        |:---|:------|
        | 1  | Cat1  |
        | 2  | Term2 |
        | 3  | Term3 |
        | 4  | Term4 |
        | 5  | Cat2  |
        | 6  | Term6 |
        | 7  | Term7 |
        | 8  | Term8 |

        To:

        | ID | Field | NewField |
        |:---|:------|:---------|
        | 2  | Term2 | Cat1     |
        | 3  | Term3 | Cat1     |
        | 4  | Term4 | Cat1     |
        | 6  | Term6 | Cat2     |
        | 7  | Term7 | Cat2     |
        | 8  | Term8 | Cat2     |
    """

    def __init__(self):
        super().__init__()
        self.name = "PIVOT_CATEGORIES"
        self.title = "Pivot categories"
        self.description = "Convert row-level categories into field categorisations."

    def parse(self, *, script: str) -> dict[str, str]:
        parsed = {}
        # Get unique assignment terms
        root = self.parser.get_split_terms(script=script, by="<")
        if len(root) != 2:
            raise ValueError(f"Category action script has no source terms ({script}).")
        # Extract source terms
        source_terms = self.parser.get_split_terms(script=root[1], by="::")
        parsed["source"] = source_terms[0]
        if len(source_terms) != 2:
            raise ValueError(f"Category pivot action script has no source category row term ({script}).")
        source_terms = self.parser.get_literal(text=source_terms[1])
        if not all([isinstance(x, int) for x in source_terms]):
            raise ValueError(f"Category pivot row terms are not integers ({script}).")
        parsed["source_category"] = source_terms
        root = self.parser.get_split_terms(script=root[0], by=">")
        # Extract destination terms
        if len(root) != 2:
            raise ValueError(f"Category action script has no destination terms ({script}).")
        parsed["destination"] = root[1]
        parsed["category"] = None
        # And the action term
        parsed["action"] = root[0]
        return parsed

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: FieldModel,
        category: None = None,
        assigned: list[int] | None = None,
    ) -> pd.DataFrame:
        # Fix: Multiple arguments results in keyerror in Modin
        df[destination.name] = None
        for i, idx in enumerate(assigned):
            to_idx = df.index[-1] + 1
            if i + 1 < len(assigned):
                to_idx = assigned[i + 1]
            category_column = {k: v for k, v in df.loc[idx].items() if pd.notnull(v) and k == source.name}
            # https://stackoverflow.com/a/46307319
            idx_list = np.arange(idx + 1, to_idx)
            # https://github.com/modin-project/modin/issues/4354
            df.loc[df.index.intersection(idx_list), destination.name] = category_column[source.name]
        return df.drop(assigned, errors="ignore")



from __future__ import annotations
from pathlib import Path

from whyqd.models import FieldModel, SchemaModel, DataSourceModel
from whyqd.crud.field import CRUDField
from whyqd.core.base import BaseDefinition


class SchemaDefinition(BaseDefinition):
    """Create and manage a metadata schema.

    !!! tip "Strategy"
        Guidance on how to use this definition is in the strategies section on
        [Schema Strategies](../strategies/schema.md).

    Parameters:
      source: A path to a json file containing a saved schema, or a dictionary conforming to the SchemaModel.

    Example:
      Create a new `SchemaDefinition` as follows:

      ```python
      import whyqd as qd

      schema: qd.models.SchemaModel = {
          "name": "urban_population",
          "title": "Urban population",
          "description": "Urban population refers to people living in urban areas as defined by national statistical offices.",
      }
      schema_destination = qd.SchemaDefinition()
      schema_destination.set(schema=schema)
      ```
    """

    def __init__(self, *, source: Path | str | SchemaModel | None = None) -> None:
        super().__init__(model_name="schema")
        self.crud = CRUDField(FieldModel)
        self.set(schema=source)

    def _refresh_model_terms(self) -> None:
        """Refreshes the list of terms and connects them to the Form Model"""
        if self.model:
            self.model.fields = self.crud.get_all()

    def _refresh_model_fields(self) -> None:
        """Refreshes the list of terms and connects them to the Field CRUD"""
        if self.model and self.model.fields:
            self.crud.reset()
            if isinstance(self.model.fields[0], dict):
                self.crud.add_multi(terms=self.model.fields)
            else:
                self.crud.add_multi(
                    terms=[f.dict(by_alias=True, exclude_defaults=True, exclude_none=True) for f in self.model.fields]
                )
        if self.model and not self.model.fields:
            self.crud.reset()

    @property
    def get(self) -> SchemaModel | None:
        """Get the schema model.

        Returns:
          Pydantic SchemaModel or None
        """
        self._refresh_model_terms()
        return self.model

    def set(self, *, schema: Path | str | SchemaModel | None = None) -> None:
        """Update or create the schema.

        Parameters:
          schema: A dictionary, or path to a dictionary or json file, conforming to the SchemaModel.
        """
        self.model = self.core.create_or_update_model(modelType=SchemaModel, source=schema, model=self.model)
        self._refresh_model_fields()

    @property
    def fields(self) -> CRUDField:
        """Returns the active crud model for all Field operations. See [Field CRUD](field.md) for API.

        Returns:
          For all Field CRUD behaviours.
        """
        return self.crud

    #########################################################################################
    # DERIVE SCHEMA FROM TABULAR DATA
    #########################################################################################

    def derive_model(self, *, data: DataSourceModel | dict) -> None:
        """Derive a schema model from a data source model.

        Arguments:
            data: A model defining a single data source
        """
        if isinstance(data, dict):
            data = DataSourceModel(**data)
        if not self.model:
            self.set()
        for dc in data.columns:
            field = dc.dict(exclude={"uuid"}, by_alias=True)
            field["title"] = field["name"]
            self.fields.add(term=field)
        self._refresh_model_terms()
        if data.index:
            self.model.index = data.index



from enum import Enum


class StatusType(str, Enum):
    WAITING = "waiting"
    PROCESSING = "processing"
    READY_MERGE = "ready_merge"
    READY_STRUCTURE = "ready_structure"
    READY_CATEGORY = "ready_categorise"
    READY_FILTER = "ready_filter"
    READY_TRANSFORM = "ready_transform"
    CREATE_ERROR = "create_error"
    MERGE_ERROR = "merge_error"
    STRUCTURE_ERROR = "structure_error"
    CATEGORY_ERROR = "category_error"
    TRANSFORM_ERROR = "transform_error"
    PROCESS_COMPLETE = "process_complete"

    @property
    def describe(self):
        description = {
            "waiting": "Waiting ...",
            "processing": "Processing ...",
            "ready_merge": "Ready to Merge",
            "ready_structure": "Ready to Structure",
            "ready_categorise": "Ready to Categorise",
            "ready_filter": "Ready to Filter",
            "ready_transform": "Ready to Transform",
            "create_error": "Create Error",
            "merge_error": "Merge Error",
            "structure_error": "Structure Error",
            "category_error": "Categorisation Error",
            "transform_error": "Transform Error",
            "process_complete": "Process Complete",
        }
        return description[self.value]



from __future__ import annotations
from pydantic import BaseModel, Field
from uuid import UUID, uuid4


class ActionScriptModel(BaseModel):
    """Action Script Model - only action `script` string required."""

    uuid: UUID = Field(default_factory=uuid4, description="Unique identity for the Action. Automatically generated.")
    script: str = Field(..., description="An action script.")

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True



from __future__ import annotations
from typing import Type, TYPE_CHECKING
import modin.pandas as pd
import numpy as np

from whyqd.parsers import CoreParser, ScriptParser  # , DataSourceParser

if TYPE_CHECKING:
    from whyqd.models import MorphActionModel, FieldModel  # , DataSourceModel
    from whyqd.core import SchemaDefinition
    from whyqd.crosswalk.base import BaseMorphAction


class MorphParser:
    """Parsing functions for morph implementations of action scripts.

    Can process and validate any morphs script. Scripts must be 'flat' and are of the form:

        "ACTION > [fields] < [rows]"

    Where:

    * the presence and order of the arrays is set by ACTION `structure`,
    * `fields` are indicated by `>`, and
    * `rows` are indicated by `<`.
    """

    def __init__(self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None) -> None:
        """
        Parameters
        ----------
        schema_source: SchemaDefinition, optional
        schema_destination: SchemaDefinition, optional
        """
        self.core = CoreParser()
        self.parser = ScriptParser()
        self.schema = []
        if schema_source and schema_destination:
            self.set_schema(schema_source=schema_source, schema_destination=schema_destination)

    ###################################################################################################
    ### PARSE TRANSFORM SCRIPT
    ###################################################################################################

    def parse(
        self,
        *,
        script: str,
        action: BaseMorphAction,
    ) -> dict[str, MorphActionModel | list[FieldModel] | list[int] | str]:
        """Generate the parsed dictionary of an initialised action script.

        Parameters
        ----------
        script: str
            An action script.
        action: BaseSchemaAction

        Raises
        ------
        ValueError for any parsing errors.

        Returns
        -------
        dict
            Parsed dictionary of an initialised action script. Of the form::
        """
        source, destination, rows, source_param, destination_param = None, None, None, None, None
        # Parse source terms
        root = self.parser.get_split_terms(script=script, by="<")
        if len(root) == 2:
            # Parsing "> 'param'::[terms]" - this is an attribute some morphs can request
            optional = self.parser.get_split_terms(script=root[1], by="::")
            if len(optional) == 2:
                # Must be a string
                source_param = self.parser.get_literal(text=optional[0])
                root[1] = optional[1]
            if "rows" in action.structure:
                # Parsing "< [rows]"
                # Must all be of type `int`
                rows = [int(i) for i in self.get_morph_struts(term=root[1])]
            if "fields" in action.structure:
                # Parsing "< [fields]"
                # Can be either of `schema_source` or `schema_destination`
                source = self.get_morph_struts(term=root[1])
                source = [self.parser.get_literal(text=f) for f in source]
                source = [self.parser.get_schema_field(term=f, schema=self.schema) for f in source]
        # Parse destination terms
        root = self.parser.get_split_terms(script=root[0], by=">")
        if len(root) == 2:
            # Parsing "> 'param'::[fields]" - this is a term some morphs can request
            optional = self.parser.get_split_terms(script=root[1], by="::")
            if len(optional) == 2:
                # Must be a string
                destination_param = self.parser.get_literal(text=optional[0])
                root[1] = optional[1]
            # Can be either of `schema_source` or `schema_destination`
            destination = self.get_morph_struts(term=root[1])
            destination = [self.parser.get_literal(text=f) for f in destination]
            destination = [self.parser.get_schema_field(term=f, schema=self.schema) for f in destination]
        # Validate structure
        if (source or destination) and "fields" not in action.structure:
            raise ValueError(f"Script error for {action.name} Morph. Fields are not a valid input.")
        if not (source or destination) and "fields" in action.structure:
            raise ValueError(f"Script error for {action.name} Morph. No valid input fields provided.")
        if rows:
            if not self.row_indices:
                raise ValueError(
                    f"Script error for {action.name} Morph. Schema does not define row indices for a row-level action."
                )
            if "rows" not in action.structure:
                raise ValueError(f"Script error for {action.name} Morph. Rows are not a valid input.")
            elif not np.in1d(rows, self.row_indices).all():
                raise ValueError("Not all row indices found in source data.")
        else:
            if "rows" in action.structure:
                raise ValueError(f"Script error for {action.name} Morph. No valid input rows provided.")
        return {
            "action": action,
            "destination": destination,
            "source": source,
            "rows": rows,
            "source_param": source_param,
            "destination_param": destination_param,
        }

    ###################################################################################################
    ### IMPLEMENT VALIDATED SCRIPT
    ###################################################################################################

    def transform(
        self,
        *,
        df: pd.DataFrame,
        action: Type[BaseMorphAction],
        source: list[FieldModel] | FieldModel | None = None,
        destination: list[FieldModel] | FieldModel | None = None,
        rows: list[int] | None = None,
        source_param: str | None = None,
        destination_param: str | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Transform a dataframe according to a morph script.

        .. warning:: Morphs can change the table columns, effecting referencing. This will be captured into the
            `DataSourceModel` and will effect Schema ACTION ColumnModel references. User beware.

        Parameters
        ----------
        action: MorphActionModel
        source: list of FieldModel or a FieldModel
            List of source fields for the action, default None.
        destination: list of FieldModel
            List of destination fields for the action, default None.
        rows: list of int, default None.
        source_param: str
            Additional parameters for a specific action, default None.
        destination_param: str
            Additional parameters for a specific action, default None.

        Returns
        -------
        Dataframe
            Containing the implementation of all transformations
        """
        return action.transform(
            df=df,
            source=source,
            destination=destination,
            rows=rows,
            source_param=source_param,
            destination_param=destination_param,
        )

    ###################################################################################################
    ### SUPPORT UTILITIES
    ###################################################################################################

    def set_schema(
        self, *, schema_source: SchemaDefinition = None, schema_destination: SchemaDefinition = None
    ) -> None:
        """Set SchemaDefinitions for the parser.

        Parameters
        ----------
        schema_source: SchemaDefinition
        schema_destination: SchemaDefinition
        """
        if not schema_source or not schema_destination:
            raise ValueError("Schema for both source and destination has not been provided.")
        self.schema = [schema_source, schema_destination]
        self.row_indices = None
        if schema_source.get.index:
            self.row_indices = list(range(schema_source.get.index))

    def get_hexed_script(self, *, script: str) -> str:
        # Changes fields to uuid hexes
        all_fields = [field for s in self.schema for field in s.get.fields]
        script = self.parser.get_hexed_script(script=script, fields=all_fields)
        return ",".join([s.strip() for s in script.split(",") if s.strip()])

    def get_morph_struts(self, *, term: str) -> list[str]:
        terms = list(self.parser.generate_contents(text=term))
        if not terms:
            # It wasn't a list
            return [term]
        if len(terms) != 1:
            raise ValueError(f"Morph actions must not be nested. ({term}).")
        return self.parser.get_split_terms(script=terms[0][1], by=",", maxsplit=-1)



from .citation import CitationModel
from .category import CategoryModel
from .column import ColumnModel
from .actionscript import ActionScriptModel
from .modifier import ModifierModel
from .version import VersionModel
from .action_schema import SchemaActionModel
from .action_morph import MorphActionModel
from .action_category import CategoryActionModel
from .constraints import ConstraintsModel
from .fields import FieldModel
from .datasource import DataSourceModel, DataSourceAttributeModel
from .schema import SchemaModel
from .crosswalk import CrosswalkModel
from .transform import TransformModel



from __future__ import annotations
from pydantic import BaseModel, Field  # , validator
from typing import List, Union, Optional, Type  # , Any

from whyqd.models.fields import FieldModel
from whyqd.models.modifier import ModifierModel


class SchemaActionModel(BaseModel):
    """Action Model - generated from the Action module."""

    name: str = Field(..., description="Name of the Action. Uppercase.")
    title: str = Field(..., description="Title of the Action. Regular case.")
    description: str = Field(..., description="Description of the purpose for performing this action.")
    structure: List[Union[str, Type[ModifierModel], Type[FieldModel]]] = Field(
        ...,
        description="The structure of an action depends on source column fields, and [optional] modifiers which act upon them.",
    )
    modifiers: Optional[List[ModifierModel]] = Field(
        None, description="The list of defined modifiers appropriate for this action."
    )

    class Config:
        use_enum_values = True
        # anystr_strip_whitespace = True
        validate_assignment = True

    # @validator("text_structure")
    # def check_valid_models(cls, v):
    #     for m in v:
    #         if not (m in ["modifier", "field", "value"]):
    #             raise ValueError(
    #                 "Structure must be of either ModifierModel or ColumnModel, or - in the case of `NEW` actions - a `value` assignment."
    #             )
    #     return v

    # @property
    # def structure(self):
    #     return [ModifierModel if s == "modifier" else ColumnModel if s == "field" else Any for s in self.text_structure]



from __future__ import annotations
from typing import TYPE_CHECKING

from whyqd.crosswalk.base import BaseSchemaAction
from whyqd.models import ModifierModel, FieldModel

if TYPE_CHECKING:
    import modin.pandas as pd


class Action(BaseSchemaAction):
    """
    Collate a list of source fields into an array of corresponding values in a destination field.

    !!! tip "Script template"
        ```python
        "COLLATE > 'destination_field' < ['source_field', 'source_field', etc.]"
        ```

        Where the order of collation is important and is used to create an array (list) of values from corresponding sources.
        Any missing values will be included as `None`.

    This can be used as part of a process which will permit multiple corresponding columns (e.g. a `date`, `description`, 
    `value` sequence) to each be placed in an array, and then pivoted using `df.explode`.

    !!! example
        ```python
        "COLLATE > 'laReliefExemptionAmount' < ['Mandatory Relief Amount', 'Discretionary Relief Amount', 'Additional Relief Amount', 'Small Business Rate Relief Amount']"
        ```
        This could be sequenced with:
        ```python
        "CATEGORISE > 'laReliefExemptionType'::'mandatory' < 'Mandatory Relief Code'::['COMMUNITY AMATEUR SPORTS CLUB RELIEFM','MANDATORY 80%']"
        "CATEGORISE > 'laReliefExemptionType'::'discretionary' < 'Discretionary Relief Code'::['20%']"
        "CATEGORISE > 'laReliefExemptionType'::'retail' < 'Additional Relief Code'::['RETAIL SCHEME']"
        "CATEGORISE > 'laReliefExemptionType'::'small_business' < 'Small Business Rate Relief'::['yes']"
        ```
        This creates two arrays, each with four elements, and permits the following `pandas` transform:
        ```python
        df = df.explode(["laReliefExemptionType","laReliefExemptionAmount"])
        ```
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = "COLLATE"
        self.title = "Collate"
        self.description = (
            "Collate a list of source fields into an array of corresponding values in a destination field."
        )
        self.structure = [ModifierModel, FieldModel]

    @property
    def modifiers(self) -> list[ModifierModel]:
        return [ModifierModel(**{"name": "~", "title": "Spacer"})]

    def validate(self, *, destination: FieldModel, source: list) -> bool:
        """
        Validate that script source structure conforms to the ACTION structure.

        Returns
        -------
        bool
        """
        if not isinstance(source, list):
            raise ValueError(f"Action source script does not conform to required structure. ({source})")
        for term in source:
            # Loops through the phrasing of the structure, and checks that each term is as expected
            # does not check that the actual terms match, though
            if isinstance(term, tuple(self.structure)):
                continue
            else:
                raise ValueError(
                    f"Source structure ({term}) doesn't conform to ACTION structure requirements ({self.structure})."
                )
        return True

    def transform(
        self,
        *,
        df: pd.DataFrame,
        destination: FieldModel,
        source: list[FieldModel | ModifierModel],
    ) -> pd.DataFrame:
        # Defaults
        # default = None
        # if destination.constraints and destination.constraints.default:
        #     default = destination.constraints.default.name
        # Array collation
        new_column = []
        if destination.name in df.columns:
            new_column = [df[destination.name].tolist()]
        spacer = [None] * len(df)
        for s in source:
            source_column = spacer
            if isinstance(s, FieldModel):
                # It's not a spacer column
                source_column = df[s.name].tolist()
            new_column.append(
                [
                    [x] if not isinstance(x, list) else x
                    for x in source_column
                ]
            )
        if len(new_column) > 1:
            # Permits reconstruction of datasets using array transformations
            # https://stackoverflow.com/a/18411610
            new_column = [[c for clist in x for c in clist] for x in zip(*new_column)]
        else:
            new_column = [[x] if not isinstance(x, list) else x for x in new_column[0]]
        df[destination.name] = new_column
        return df



from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
from pydantic import Json
import json

from whyqd.dtypes import MimeType
from whyqd.models import DataSourceModel, CitationModel
from whyqd.parsers import DataSourceParser
from whyqd.core.base import BaseDefinition

if TYPE_CHECKING:
    import modin.pandas as pd


class DataSourceDefinition(BaseDefinition):
    """Create and manage a data source schema.

    !!! tip "Strategy"
        Guidance on how to use this definition is in the strategies section on
        [DataSource Strategies](../strategies/datasource.md).

    !!! info
        **whyqd** supports any of the following file mime types:

        - `CSV`: "text/csv"
        - `XLS`: "application/vnd.ms-excel"
        - `XLSX`: "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        - `PARQUET` (or `PRQ`): "application/vnd.apache.parquet"
        - `FEATHER` (or `FTR`): "application/vnd.apache.feather"

        Declare it like so:

        ```python
        MIMETYPE = "xlsx" # upper- or lower-case is fine
        ```

        Specify the mime type as a text string, uppper- or lower-case. Neither of
        [Parquet](https://parquet.apache.org/docs/overview/) or [Feather](https://arrow.apache.org/docs/python/feather.html)
        yet have official mimetypes, so this is what we're using for now.

    Parameters:
      source: A path to a json file containing a saved schema, or a dictionary conforming to the DataSourceModel.

    Example:
      Create and validate a new `DataSourceDefinition` as follows:

      ```python
      import whyqd as qd

      datasource = qd.DataSourceDefinition()
      datasource.derive_model(source=DATASOURCE_PATH, mimetype=MIMETYPE)
      datasource.save(directory=DIRECTORY)
      datasource.validate()
      ```
    """

    def __init__(self, *, source: Path | str | DataSourceModel | None = None) -> None:
        super().__init__(model_name="data")
        self.reader = DataSourceParser()
        self.data = None
        self.set(schema=source)

    @property
    def get(self) -> DataSourceModel | list[DataSourceModel] | None:
        """Get the data source model.

        !!! warning
            If your source data are `Excel`, and that spreadsheet consists of multiple `sheets`, then **whyqd** will
            produce multiple data models which will be returned as a list. Each model will reflect the metadata for
            each sheet.

            As always *look* at your data and test before implementing in code. You should see an additional `sheet_name`
            field.

        Returns:
          Pydantic DataSourceModel as a list, a single, or None
        """
        return self.model

    def set(self, *, schema: Path | str | DataSourceModel | None = None) -> None:
        """Update or create the schema.

        Parameters:
          schema: A dictionary conforming to the DataSourceModel.
        """
        if schema:
            self.model = self.core.create_or_update_model(modelType=DataSourceModel, source=schema, model=self.model)

    def get_data(self, *, refresh: bool = False) -> pd.DataFrame | None:
        """Get a Pandas (Modin) dataframe.

        Parameters:
          refresh: Force an update of the dataframe if there have been attribute changes.

        Returns:
          A dataframe, or none.
        """
        if self.model and (refresh or self.data is None or self.data.empty):
            self.data = self.reader.get(source=self.model)
        return self.data

    #########################################################################################
    # DERIVE SCHEMA FROM TABULAR DATA
    #########################################################################################

    def derive_model(
        self,
        *,
        source: Path | str,
        mimetype: str | MimeType,
        header: int | list[int | None] | None = 0,
        **attributes,
    ) -> DataSourceModel | list[DataSourceModel]:
        """Derive a data model schema (or list) from a data source. All columns will be coerced to `string` type to
        preserve data quality even though this is far less efficient.

        Parameters:
          source:  Source filename.
          mimetype:  Pandas can read a diversity of mimetypes. **whyqd** supports `xls`, `xlsx`, `csv`, `parquet` and `feather`.
          header:  Row (0-indexed) to use for the column labels of the parsed DataFrame. If there are multiple sheets, then
                    a list of integers should be provided. If `header` is `None`, row 0 will be treated as values and a
                    set of field names will be generated indexed to the number of data columns.
          attributes: dict of specific `mimetype` related Pandas attributes. Use sparingly.

        Returns:
          List of DataSourceModel, or DataSourceModel
        """
        self.model = self.reader.derive_data_model(source=source, mimetype=mimetype, header=header, **attributes)

    #########################################################################################
    # VALIDATION UTILITIES
    #########################################################################################

    def validate(self) -> bool:
        """Validate that all required fields are returned from the crosswalk."""
        if not self.model:
            raise ValueError(f"{self.model_name.title()} model does not exist.")
        models = self.model
        if not isinstance(self.model, list):
            models = [models]
        for m in models:
            attributes = {}
            if m.attributes:
                attributes |= m.attributes.terms
            vmod = self.reader.derive_data_model(
                source=m.path, mimetype=m.mime, sheet_name=m.sheet_name, header=m.header, **attributes
            )
            try:
                assert vmod.checksum == m.checksum
            except AssertionError:
                raise AssertionError(f"{self.model_name.title()} model checksum does not match derived checksum.")
        return True

    #########################################################################################
    # MANAGE CITATION
    #########################################################################################

    def get_citation(self) -> dict[str, str | dict[str, str]]:
        """Get the citation as a dictionary.

        Raises:
          ValueError: If no citation has been declared or the build is incomplete.

        Returns:
          A dictionary conforming to the CitationModel.
        """
        if not isinstance(self.model, list):
            return super().get_citation()
        return [
            m.citation.dict(by_alias=True, exclude_defaults=True, exclude_none=True) for m in self.model if m.citation
        ]

    def set_citation(self, *, citation: CitationModel, index: int = None) -> None:
        """Update or create the citation.

        Parameters:
          citation: A dictionary conforming to the CitationModel.
          index: If there are multiple sources from the source data, provide the index (base 0) for the resource citation.
        """
        # Create a temporary CitationModel
        if not isinstance(self.model, list):
            super().set_citation(citation=citation)
        elif index:
            self.model[index].citation = self.core.create_or_update_model(
                modelType=CitationModel, source=citation, model=self.model[index].citation
            )
        else:
            for index in range(len(self.model)):
                self.model[index].citation = self.core.create_or_update_model(
                    modelType=CitationModel, source=citation, model=self.model[index].citation
                )

    #########################################################################################
    # SAVE UTILITIES FOR LISTS
    #########################################################################################

    def get_json(self, hide_uuid: bool = False) -> Json | None:
        """Get the json model.

        Parameters:
          hide_uuid: Hide all UUIDs in the nested JSON output. Mostly useful for validation assertions where the only
                     differences between sources are the UUIDs.

        Returns:
          Json-conforming output, or None.
        """
        if not isinstance(self.model, list):
            return super().get_json(hide_uuid=hide_uuid)
        response = []
        for model in self.model:
            if not hide_uuid:
                response.append(model.json(by_alias=True, exclude_defaults=True, exclude_none=True))
            else:
                json.dumps(self.exclude_uuid(model=model))
        return response

    def save(
        self,
        directory: str | None = None,
        filename: str | None = None,
        created_by: str | None = None,
        hide_uuid: bool = False,
    ) -> bool:
        """Save model as a json file.

        Parameters:
          directory:  Defaults to working directory
          filename:  Defaults to model name
          created_by:  Declare the model creator/updater
          hide_uuid:  Hide all UUIDs in the nested JSON output.

        Returns:
          Boolean True if saved.
        """
        if not isinstance(self.model, list):
            if not filename:
                filename = "".join(self.model.name.split(".")[:-1] + [f".{self.model_name}"])
            return super().save(directory=directory, filename=filename, created_by=created_by, hide_uuid=hide_uuid)
        if not directory:
            directory = self.core.default_directory
        else:
            directory = self.core.check_path(directory=directory)
        if not filename:
            filename = f"{self.model.name}.{self.model_name}"
        if filename.split(".")[-1] != self.model_name:
            filename += f".{self.model_name}"
        if isinstance(directory, str):
            directory = Path(directory)
        models = self.get_json(hide_uuid=hide_uuid)
        for i, model in enumerate(models):
            filename = "".join(self.model[i].name.split(".")[:-1] + [f".{self.model_name}"])
            path = directory / filename
            self.core.save_file(data=model, source=path)
        return True
